"""PhotoLibrary tests."""

# pylint: disable=protected-access
import base64
from datetime import datetime, timezone
from typing import Any
from unittest.mock import MagicMock, mock_open, patch

import pytest

from pyicloud.const import CONTENT_TYPE, CONTENT_TYPE_TEXT
from pyicloud.exceptions import (
    PyiCloudAPIResponseException,
    PyiCloudServiceNotActivatedException,
)
from pyicloud.services.photos import (
    PRIMARY_ZONE,
    AlbumContainer,
    AlbumTypeEnum,
    BasePhotoAlbum,
    BasePhotoLibrary,
    DirectionEnum,
    ListTypeEnum,
    ObjectTypeEnum,
    PhotoAlbum,
    PhotoAsset,
    PhotoLibrary,
    PhotosService,
    PhotoStreamLibrary,
    SharedPhotoStreamAlbum,
    SmartAlbumEnum,
)


def test_photo_library_initialization(mock_photos_service: MagicMock) -> None:
    """Tests initialization of PhotoLibrary."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
    ]
    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    assert library.zone_id == {"zoneName": "PrimarySync"}
    assert library.url == ("https://example.com/records/query?dsid=12345")


def test_photo_library_indexing_not_finished(mock_photos_service: MagicMock) -> None:
    """Tests exception when indexing is not finished."""
    mock_photos_service.session.post.return_value.json.return_value = {
        "records": [
            {
                "fields": {
                    "state": {"value": "NOT_FINISHED"},
                },
            }
        ]
    }
    with pytest.raises(PyiCloudServiceNotActivatedException):
        PhotoLibrary(
            service=mock_photos_service,
            zone_id={"zoneName": "PrimarySync"},
            upload_url="https://upload.example.com",
        )


def test_fetch_folders(mock_photos_service: MagicMock) -> None:
    """Tests the _fetch_folders method."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder1",
                            "recordChangeTag": "tag1",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMQ=="},
                                "isDeleted": {"value": False},
                            },
                        }
                    ]
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    library.SMART_ALBUMS = {}
    albums: AlbumContainer = library.albums

    assert len(albums) == 1
    assert albums[0].name == "folder1"


def test_get_albums(mock_photos_service: MagicMock) -> None:
    """Tests the _get_albums method."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder1",
                            "recordChangeTag": "tag1",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMQ=="},
                                "isDeleted": {"value": False},
                            },
                        },
                        {
                            "recordName": "1111-1111-1111-1111",
                            "recordChangeTag": "tag2",
                            "fields": {
                                "albumNameEnc": {"value": "QWxidW0gTmFtZSAy"},
                                "isDeleted": {"value": False},
                            },
                        },
                    ]
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    albums: AlbumContainer = library.albums
    assert SmartAlbumEnum.ALL_PHOTOS in albums
    assert "folder1" in albums
    assert albums["folder1"].name == "folder1"
    assert albums["Album Name 2"].id == "1111-1111-1111-1111"
    assert albums.index(1).id == "Time-lapse"
    assert albums.get("Nonexistent Album") is None
    assert albums[0] == next(iter(albums))
    with pytest.raises(KeyError):
        _ = albums["Album Name 3"]

    with pytest.raises(IndexError):
        _ = albums.index(100)


def test_upload_file_success(mock_photos_service: MagicMock) -> None:
    """Tests the upload_file method for successful upload."""
    mock_photos_service.params = {"dsid": "12345"}
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "uploaded_photo",
                            "recordChangeTag": "tag1",
                            "recordType": "CPLAsset",
                            "fields": {
                                "masterRef": {"value": {"recordName": "uploaded_photo"}}
                            },
                        },
                        {
                            "recordType": "CPLMaster",
                            "recordName": "uploaded_photo",
                        },
                    ]
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )

    with patch("builtins.open", mock_open(read_data=b"file_content")) as mock_file:
        asset: PhotoAsset | None = library.upload_file("test_photo.jpg")

    assert asset is not None
    assert asset.id == "uploaded_photo"
    mock_photos_service.session.post.assert_called_with(
        url="https://upload.example.com/upload?dsid=12345&filename=test_photo.jpg",
        data=mock_file.return_value,
    )


def test_upload_file_with_errors(mock_photos_service: MagicMock) -> None:
    """Tests the upload_file method when the response contains errors."""
    mock_photos_service.params = {"dsid": "12345"}
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "errors": [
                        {
                            "code": "UPLOAD_ERROR",
                            "message": "Upload failed",
                        },
                    ],
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )

    with patch("builtins.open", mock_open(read_data=b"file_content")) as mock_file:
        with pytest.raises(PyiCloudAPIResponseException) as exc_info:
            library.upload_file("test_photo.jpg")

    assert "UPLOAD_ERROR" in str(exc_info.value)
    mock_photos_service.session.post.assert_called_with(
        url="https://upload.example.com/upload?dsid=12345&filename=test_photo.jpg",
        data=mock_file.return_value,
    )


def test_upload_file_no_records(mock_photos_service: MagicMock) -> None:
    """Tests the upload_file method when no records are returned."""
    mock_photos_service.params = {"dsid": "12345"}
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [],
                }
            )
        ),
    ]
    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )

    with patch("builtins.open", mock_open(read_data=b"file_content")) as mock_file:
        result: PhotoAsset | None = library.upload_file("test_photo.jpg")
        assert result is None

    mock_photos_service.session.post.assert_called_with(
        url="https://upload.example.com/upload?dsid=12345&filename=test_photo.jpg",
        data=mock_file.return_value,
    )


def test_fetch_folders_multiple_pages(mock_photos_service: MagicMock) -> None:
    """Tests _fetch_folders with multiple pages of results."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        },
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder1",
                            "recordChangeTag": "tag1",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMQ=="},
                                "isDeleted": {"value": False},
                            },
                        }
                    ],
                    "continuationMarker": "marker1",
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder2",
                            "recordChangeTag": "tag2",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMg=="},
                                "isDeleted": {"value": False},
                            },
                        }
                    ]
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    library.SMART_ALBUMS = {}
    albums: AlbumContainer = library.albums
    assert len(albums) == 2
    assert albums[0].name == "folder1"
    assert albums[1].name == "folder2"
    mock_photos_service.session.post.assert_called()


def test_fetch_folders_skips_deleted_folders(mock_photos_service: MagicMock) -> None:
    """Tests _fetch_folders skips folders marked as deleted."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        },
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "continuationMarker": "marker1",
                    "records": [
                        {
                            "recordName": "folder1",
                            "recordChangeTag": "tag1",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMQ=="},
                                "isDeleted": {"value": True},
                            },
                        },
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder2",
                            "recordChangeTag": "tag2",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMg=="},
                                "isDeleted": {"value": False},
                            },
                        },
                    ]
                },
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    library.SMART_ALBUMS = {}
    albums: AlbumContainer = library.albums

    assert len(albums) == 1
    assert albums[0].name == "folder2"
    mock_photos_service.session.post.assert_called()


def test_fetch_folders_no_records(mock_photos_service: MagicMock) -> None:
    """Tests _fetch_folders when no records are returned."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        },
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [],
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    library.SMART_ALBUMS = {}
    albums: AlbumContainer = library.albums

    assert len(albums) == 0
    mock_photos_service.session.post.assert_called()


def test_fetch_folders_handles_missing_fields(mock_photos_service: MagicMock) -> None:
    """Tests _fetch_folders handles records with missing fields."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        },
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder1",
                            "fields": {
                                "isDeleted": {"value": False},
                            },
                        }
                    ]
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    library.SMART_ALBUMS = {}
    albums: AlbumContainer = library.albums

    assert len(albums) == 0
    mock_photos_service.session.post.assert_called()


def test_base_photo_album_initialization(mock_photo_library: MagicMock) -> None:
    """Tests initialization of BasePhotoAlbum."""
    album = BasePhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        list_type=ListTypeEnum.DEFAULT,
        page_size=50,
        direction=DirectionEnum.ASCENDING,
    )
    assert album.name == "Test Album"
    assert album.service == mock_photo_library.service
    assert album.page_size == 50


def test_base_photo_album_parse_response() -> None:
    """Tests the _parse_response method."""
    library = BasePhotoLibrary(
        service=MagicMock(),
        asset_type=PhotoAsset,
    )
    response = {
        "records": [
            {
                "recordType": "CPLAsset",
                "fields": {"masterRef": {"value": {"recordName": "master1"}}},
            },
            {
                "recordType": "CPLMaster",
                "recordName": "master1",
            },
        ]
    }
    asset_records, master_records = library.parse_asset_response(response)
    assert "master1" in asset_records
    assert len(master_records) == 1
    assert master_records[0]["recordName"] == "master1"


def test_base_photo_album_get_photos_at(mock_photo_library: MagicMock) -> None:
    """Tests the _get_photos_at method."""
    mock_photo_library.service.session.post.return_value.json.side_effect = [
        {
            "records": [
                {
                    "recordType": "CPLAsset",
                    "fields": {"masterRef": {"value": {"recordName": "master1"}}},
                },
                {
                    "recordType": "CPLMaster",
                    "recordName": "master1",
                },
            ]
        },
        {
            "records": [],
        },
    ]
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        list_type=ListTypeEnum.DEFAULT,
        obj_type=ObjectTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        page_size=10,
        record_id="album1",
        url="https://example.com/records/query?dsid=12345",
    )
    photos = list(album.photos)
    assert len(photos) == 1
    mock_photo_library.service.session.post.assert_called()


def test_base_photo_album_len(mock_photos_service: MagicMock) -> None:
    """Tests the __len__ method."""
    album = BasePhotoAlbum(
        library=mock_photos_service,
        name="Test Album",
        list_type=ListTypeEnum.DEFAULT,
    )
    album._get_len = MagicMock(return_value=42)
    assert len(album) == 42
    album._get_len.assert_called_once()


def test_base_photo_album_iter(mock_photo_library: MagicMock) -> None:
    """Tests the __iter__ method."""
    mock_photo_library.service.session.post.return_value.json.side_effect = [
        {
            "records": [
                {
                    "recordType": "CPLAsset",
                    "fields": {"masterRef": {"value": {"recordName": "master1"}}},
                },
                {
                    "recordType": "CPLMaster",
                    "recordName": "master1",
                },
            ]
        },
        {
            "records": [],
        },
    ]
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        list_type=ListTypeEnum.DEFAULT,
        obj_type=ObjectTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        page_size=10,
        url="https://example.com/records/query?dsid=12345",
        record_id="album1",
    )
    photos = list(iter(album))
    assert len(photos) == 1
    mock_photo_library.service.session.post.assert_called()


def test_base_photo_album_str() -> None:
    """Tests the __str__ method."""
    album = BasePhotoAlbum(
        library=MagicMock(),
        name="Test Album",
        list_type=ListTypeEnum.DEFAULT,
    )
    assert str(album) == "Test Album"


def test_base_photo_album_repr() -> None:
    """Tests the __repr__ method."""
    album = BasePhotoAlbum(
        library=MagicMock(),
        name="Test Album",
        list_type=ListTypeEnum.DEFAULT,
    )
    assert repr(album) == "<BasePhotoAlbum: 'Test Album'>"


def test_photos_service_initialization(mock_photos_service: MagicMock) -> None:
    """Tests initialization of PhotosService."""
    mock_photos_service.session.post.return_value.json.return_value = {
        "records": [
            {
                "fields": {
                    "state": {"value": "FINISHED"},
                },
            }
        ]
    }
    photos_service = PhotosService(
        service_root="https://example.com",
        session=mock_photos_service.session,
        params={"dsid": "12345"},
        upload_url="https://upload.example.com",
        shared_streams_url="https://shared.example.com",
    )
    assert photos_service.service_endpoint == (
        "https://example.com/database/1/com.apple.photos.cloud/production/private"
    )
    assert isinstance(photos_service._root_library, PhotoLibrary)
    assert isinstance(photos_service._shared_library, PhotoStreamLibrary)
    assert photos_service.params["remapEnums"] is True
    assert photos_service.params["getCurrentSyncToken"] is True


def test_photos_service_libraries(mock_photos_service: MagicMock) -> None:
    """Tests the libraries property."""
    mock_photos_service.session.post.return_value.json.side_effect = [
        {
            "records": [
                {
                    "fields": {
                        "state": {"value": "FINISHED"},
                    },
                }
            ]
        },
        {
            "zones": [
                {"zoneID": {"zoneName": "CustomZone"}, "deleted": False},
            ]
        },
        {
            "records": [
                {
                    "fields": {
                        "state": {"value": "FINISHED"},
                    },
                }
            ]
        },
    ]
    photos_service = PhotosService(
        service_root="https://example.com",
        session=mock_photos_service.session,
        params={"dsid": "12345"},
        upload_url="https://upload.example.com",
        shared_streams_url="https://shared.example.com",
    )
    libraries: dict[str, BasePhotoLibrary] = photos_service.libraries
    assert "root" in libraries
    assert "shared" in libraries
    assert "CustomZone" in libraries
    assert isinstance(libraries["root"], PhotoLibrary)
    assert isinstance(libraries["shared"], PhotoStreamLibrary)
    assert isinstance(libraries["CustomZone"], PhotoLibrary)
    mock_photos_service.session.post.assert_called_with(
        url=(
            "https://example.com/database/1/com.apple.photos.cloud/production/private/records/query"
            "?dsid=12345&remapEnums=True&getCurrentSyncToken=True"
        ),
        json={
            "query": {"recordType": "CheckIndexingState"},
            "zoneID": {"zoneName": "CustomZone"},
        },
        headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
    )


def test_photos_service_libraries_cached(mock_photos_service: MagicMock) -> None:
    """Tests that libraries are cached after the first access."""
    mock_photos_service.session.post.return_value.json.return_value = {
        "records": [
            {
                "fields": {
                    "state": {"value": "FINISHED"},
                },
            }
        ]
    }
    photos_service = PhotosService(
        service_root="https://example.com",
        session=mock_photos_service.session,
        params={"dsid": "12345"},
        upload_url="https://upload.example.com",
        shared_streams_url="https://shared.example.com",
    )
    mock_libraries = {"cached": MagicMock(spec=PhotoLibrary)}
    photos_service._libraries = mock_libraries  # type: ignore
    libraries: dict[str, BasePhotoLibrary] = photos_service.libraries
    assert libraries == mock_libraries
    mock_photos_service.session.post.assert_called_once()


def test_photos_service_albums(mock_photos_service: MagicMock) -> None:
    """Tests the albums property."""
    mock_photos_service.session.post.return_value.json.return_value = {
        "records": [
            {
                "fields": {
                    "state": {"value": "FINISHED"},
                },
            }
        ]
    }
    photos_service = PhotosService(
        service_root="https://example.com",
        session=mock_photos_service.session,
        params={"dsid": "12345"},
        upload_url="https://upload.example.com",
        shared_streams_url="https://shared.example.com",
    )
    albums: AlbumContainer = photos_service.albums
    assert isinstance(albums, AlbumContainer)
    assert SmartAlbumEnum.ALL_PHOTOS in albums
    mock_photos_service.session.post.assert_called()


def test_photos_service_shared_streams(mock_photos_service: MagicMock) -> None:
    """Tests the shared_streams property."""
    mock_photos_service.session.post.return_value.json.side_effect = [
        {
            "records": [
                {
                    "fields": {
                        "state": {"value": "FINISHED"},
                    },
                }
            ]
        },
        {
            "albums": [
                {
                    "albumlocation": "https://shared.example.com/album/",
                    "albumctag": "ctag",
                    "albumguid": "guid",
                    "ownerdsid": "owner",
                    "attributes": {
                        "name": "Shared Album",
                        "creationDate": "1234567890",
                        "allowcontributions": True,
                        "ispublic": False,
                    },
                    "sharingtype": "owned",
                    "iswebuploadsupported": True,
                }
            ]
        },
    ]
    photos_service = PhotosService(
        service_root="https://example.com",
        session=mock_photos_service.session,
        params={"dsid": "12345"},
        upload_url="https://upload.example.com",
        shared_streams_url="https://shared.example.com",
    )
    shared_streams: AlbumContainer = photos_service.shared_streams
    assert isinstance(shared_streams, AlbumContainer)
    assert "Shared Album" in shared_streams
    assert isinstance(shared_streams.find("Shared Album"), SharedPhotoStreamAlbum)
    mock_photos_service.session.post.assert_called()


def test_photo_album_initialization(mock_photo_library: MagicMock) -> None:
    """Tests initialization of PhotoAlbum."""
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
        query_filter=[
            {
                "fieldName": "test",
                "comparator": "EQUALS",
                "fieldValue": {"type": "STRING", "value": "test"},
            }
        ],
        zone_id={"zoneName": "TestZone"},
        page_size=50,
        parent_id="parent123",
        record_change_tag="tag123",
        record_modification_date="2023-01-01T00:00:00Z",
    )

    assert album.name == "Test Album"
    assert album.id == "album123"
    assert album._record_id == "album123"
    assert album._obj_type == ObjectTypeEnum.CONTAINER
    assert album._list_type == ListTypeEnum.CONTAINER
    assert album._direction == DirectionEnum.ASCENDING
    assert album._url == "https://example.com/records/query?dsid=12345"
    assert album._parent_id == "parent123"
    assert album._record_change_tag == "tag123"
    assert album._record_modification_date == "2023-01-01T00:00:00Z"
    assert album._zone_id == {"zoneName": "TestZone"}


def test_photo_album_initialization_default_zone(mock_photo_library: MagicMock) -> None:
    """Tests PhotoAlbum initialization with default zone."""
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
    )

    assert album._zone_id == PRIMARY_ZONE


def test_photo_album_fullname_no_parent(mock_photo_library: MagicMock) -> None:
    """Tests fullname property when album has no parent."""
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Root Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
    )

    assert album.fullname == "Root Album"


def test_photo_album_fullname_with_parent() -> None:
    """Tests fullname property when album has a parent."""
    mock_photo_library: MagicMock = MagicMock(spec=PhotoLibrary)
    parent_album = MagicMock()
    parent_album.fullname = "Parent Album"

    mock_albums = MagicMock()
    mock_albums.__getitem__.return_value = parent_album
    mock_photo_library.albums = mock_albums

    album = PhotoAlbum(
        library=mock_photo_library,
        name="Child Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
        parent_id="parent123",
    )

    assert album.fullname == "Parent Album/Child Album"
    mock_albums.__getitem__.assert_called_once_with("parent123")


def test_photo_album_rename_success(mock_photos_service: MagicMock) -> None:
    """Tests successful album renaming."""
    mock_photo_library: MagicMock = MagicMock(spec=PhotoLibrary)
    mock_photo_library.service = mock_photos_service
    mock_photo_library.service.session.post.return_value = MagicMock()
    mock_photo_library.service.service_endpoint = "https://example.com/endpoint"
    mock_photo_library.service.params = {"dsid": "12345"}

    album = PhotoAlbum(
        library=mock_photo_library,
        name="Old Name",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
        record_change_tag="tag123",
        zone_id={"zoneName": "TestZone"},
    )

    album.rename("New Name")

    assert album._name == "New Name"

    expected_data = {
        "atomic": True,
        "zoneID": {"zoneName": "TestZone"},
        "operations": [
            {
                "operationType": "update",
                "record": {
                    "recordName": "album123",
                    "recordType": "CPLAlbum",
                    "recordChangeTag": "tag123",
                    "fields": {
                        "albumNameEnc": {
                            "value": base64.b64encode(
                                "New Name".encode("utf-8")
                            ).decode("utf-8"),
                        },
                    },
                },
            }
        ],
    }

    mock_photo_library.service.session.post.assert_called_once_with(
        "https://example.com/endpoint/records/modify?dsid=12345",
        json=expected_data,
        headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
    )

    # Verify that if the server returns updated tags, they are stored
    mock_photo_library.service.session.post.return_value.json.return_value = {
        "records": [
            {
                "recordChangeTag": "new_tag",
                "fields": {"recordModificationDate": {"value": "2023-02-01T00:00:00Z"}},
            }
        ]
    }
    album.rename("Another Name")
    assert album._record_change_tag == "new_tag"
    assert album._record_modification_date == "2023-02-01T00:00:00Z"


def test_photo_album_rename_same_name(mock_photo_library: MagicMock) -> None:
    """Tests that renaming to the same name does nothing."""
    mock_photo_library.service.session.post.return_value = MagicMock()

    album = PhotoAlbum(
        library=mock_photo_library,
        name="Same Name",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
    )

    album.rename("Same Name")

    assert album._name == "Same Name"
    mock_photo_library.service.session.post.assert_not_called()


def test_photo_album_delete_success(mock_photo_library: MagicMock) -> None:
    """Tests successful album deletion."""
    mock_photo_library.service.session.post.return_value = MagicMock()
    mock_photo_library.service.service_endpoint = "https://example.com/endpoint"
    mock_photo_library.service.params = {"dsid": "12345"}

    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
        record_change_tag="tag123",
        zone_id={"zoneName": "TestZone"},
    )

    result = album.delete()

    assert result is True

    expected_data = {
        "atomic": True,
        "zoneID": {"zoneName": "TestZone"},
        "operations": [
            {
                "operationType": "update",
                "record": {
                    "recordName": "album123",
                    "recordChangeTag": "tag123",
                    "recordType": "CPLAlbum",
                    "fields": {
                        "isDeleted": {"value": 1},
                    },
                },
            }
        ],
    }

    mock_photo_library.service.session.post.assert_called_once_with(
        "https://example.com/endpoint/records/modify?dsid=12345",
        json=expected_data,
        headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
    )


def test_photo_album_upload_success(mock_photos_service: MagicMock) -> None:
    """Tests successful photo upload to album."""
    mock_photo_library: MagicMock = MagicMock(spec=PhotoLibrary)
    mock_photo_asset = MagicMock()
    mock_photo_asset.id = "photo123"
    mock_photo_library.service = mock_photos_service
    mock_photo_library.upload_file.return_value = mock_photo_asset
    mock_photo_library.service.session.post.return_value = MagicMock()
    mock_photo_library.service.service_endpoint = "https://example.com/endpoint"
    mock_photo_library.service.params = {"dsid": "12345"}

    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
        zone_id={"zoneName": "TestZone"},
    )

    result = album.upload("/path/to/photo.jpg")

    assert result == mock_photo_asset
    mock_photo_library.upload_file.assert_called_once_with("/path/to/photo.jpg")

    expected_data = {
        "atomic": True,
        "zoneID": {"zoneName": "TestZone"},
        "operations": [
            {
                "operationType": "create",
                "record": {
                    "fields": {
                        "itemId": {"value": "photo123"},
                        "position": {"value": 1024},
                        "containerId": {"value": "album123"},
                    },
                    "recordType": "CPLContainerRelation",
                    "recordName": "photo123-IN-album123",
                },
            }
        ],
    }

    mock_photo_library.service.session.post.assert_called_with(
        "https://example.com/endpoint/records/modify?dsid=12345",
        json=expected_data,
        headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
    )


def test_photo_album_upload_not_photo_library() -> None:
    """Tests upload when library is not a PhotoLibrary instance."""
    mock_library = MagicMock(spec=BasePhotoLibrary)

    album = PhotoAlbum(
        library=mock_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
    )

    result = album.upload("/path/to/photo.jpg")

    assert result is None


def test_photo_album_upload_upload_file_returns_none() -> None:
    """Tests upload when upload_file returns None."""
    mock_photo_library = MagicMock(spec=PhotoLibrary)
    mock_photo_library.upload_file.return_value = None
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
    )

    result: PhotoAsset | None = album.upload("/path/to/photo.jpg")

    mock_photo_library.upload_file.assert_called_once_with("/path/to/photo.jpg")
    assert result is None


def test_photo_album_get_container_id(mock_photo_library: MagicMock) -> None:
    """Tests _get_container_id property."""
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
    )

    container_id = album._get_container_id

    assert container_id == f"{ObjectTypeEnum.CONTAINER.value}:album123"


def test_photo_album_get_len(mock_photo_library: MagicMock) -> None:
    """Tests _get_len method."""
    mock_response = {"batch": [{"records": [{"fields": {"itemCount": {"value": 42}}}]}]}
    mock_photo_library.service.session.post.return_value.json.return_value = (
        mock_response
    )
    mock_photo_library.service.service_endpoint = "https://example.com/endpoint"
    mock_photo_library.service.params = {"dsid": "12345"}

    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
        zone_id={"zoneName": "TestZone"},
    )

    length = album._get_len()

    assert length == 42

    expected_json = {
        "batch": [
            {
                "resultsLimit": 1,
                "query": {
                    "recordType": "HyperionIndexCountLookup",
                    "filterBy": {
                        "fieldName": "indexCountID",
                        "comparator": "IN",
                        "fieldValue": {
                            "type": "STRING_LIST",
                            "value": [f"{ObjectTypeEnum.CONTAINER.value}:album123"],
                        },
                    },
                },
                "zoneWide": True,
                "zoneID": {"zoneName": "TestZone"},
            }
        ]
    }

    mock_photo_library.service.session.post.assert_called_once_with(
        "https://example.com/endpoint/internal/records/query/batch?dsid=12345",
        json=expected_json,
        headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
    )


def test_photo_album_get_payload(mock_photo_library: MagicMock) -> None:
    """Tests _get_payload method."""
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
        query_filter=[
            {
                "fieldName": "test",
                "comparator": "EQUALS",
                "fieldValue": {"type": "STRING", "value": "test"},
            }
        ],
        zone_id={"zoneName": "TestZone"},
    )

    payload = album._get_payload(
        offset=10, page_size=20, direction=DirectionEnum.DESCENDING
    )

    expected_payload = {
        "query": {
            "recordType": ListTypeEnum.CONTAINER.value,
            "filterBy": [
                {
                    "fieldName": "direction",
                    "comparator": "EQUALS",
                    "fieldValue": {
                        "type": "STRING",
                        "value": DirectionEnum.DESCENDING.value,
                    },
                },
                {
                    "fieldName": "startRank",
                    "comparator": "EQUALS",
                    "fieldValue": {"type": "INT64", "value": 10},
                },
                {
                    "fieldName": "test",
                    "comparator": "EQUALS",
                    "fieldValue": {"type": "STRING", "value": "test"},
                },
            ],
        },
        "resultsLimit": 20,
        "desiredKeys": [
            "resJPEGFullWidth",
            "resJPEGFullHeight",
            "resJPEGFullFileType",
            "resJPEGFullFingerprint",
            "resJPEGFullRes",
            "resJPEGLargeWidth",
            "resJPEGLargeHeight",
            "resJPEGLargeFileType",
            "resJPEGLargeFingerprint",
            "resJPEGLargeRes",
            "resJPEGMedWidth",
            "resJPEGMedHeight",
            "resJPEGMedFileType",
            "resJPEGMedFingerprint",
            "resJPEGMedRes",
            "resJPEGThumbWidth",
            "resJPEGThumbHeight",
            "resJPEGThumbFileType",
            "resJPEGThumbFingerprint",
            "resJPEGThumbRes",
            "resVidFullWidth",
            "resVidFullHeight",
            "resVidFullFileType",
            "resVidFullFingerprint",
            "resVidFullRes",
            "resVidMedWidth",
            "resVidMedHeight",
            "resVidMedFileType",
            "resVidMedFingerprint",
            "resVidMedRes",
            "resVidSmallWidth",
            "resVidSmallHeight",
            "resVidSmallFileType",
            "resVidSmallFingerprint",
            "resVidSmallRes",
            "resSidecarWidth",
            "resSidecarHeight",
            "resSidecarFileType",
            "resSidecarFingerprint",
            "resSidecarRes",
            "itemType",
            "dataClassType",
            "filenameEnc",
            "originalOrientation",
            "resOriginalWidth",
            "resOriginalHeight",
            "resOriginalFileType",
            "resOriginalFingerprint",
            "resOriginalRes",
            "resOriginalAltWidth",
            "resOriginalAltHeight",
            "resOriginalAltFileType",
            "resOriginalAltFingerprint",
            "resOriginalAltRes",
            "resOriginalVidComplWidth",
            "resOriginalVidComplHeight",
            "resOriginalVidComplFileType",
            "resOriginalVidComplFingerprint",
            "resOriginalVidComplRes",
            "isDeleted",
            "isExpunged",
            "dateExpunged",
            "remappedRef",
            "recordName",
            "recordType",
            "recordChangeTag",
            "masterRef",
            "adjustmentRenderType",
            "assetDate",
            "addedDate",
            "isFavorite",
            "isHidden",
            "orientation",
            "duration",
            "assetSubtype",
            "assetSubtypeV2",
            "assetHDRType",
            "burstFlags",
            "burstFlagsExt",
            "burstId",
            "captionEnc",
            "locationEnc",
            "locationV2Enc",
            "locationLatitude",
            "locationLongitude",
            "adjustmentType",
            "timeZoneOffset",
            "vidComplDurValue",
            "vidComplDurScale",
            "vidComplDispValue",
            "vidComplDispScale",
            "vidComplVisibilityState",
            "customRenderedValue",
            "containerId",
            "itemId",
            "position",
            "isKeyAsset",
        ],
        "zoneID": {"zoneName": "TestZone"},
    }

    assert payload == expected_payload


def test_photo_album_get_payload_no_query_filter(mock_photo_library: MagicMock) -> None:
    """Tests _get_payload method without query filter."""
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
        zone_id={"zoneName": "TestZone"},
    )

    payload: dict[str, Any] = album._get_payload(
        offset=5, page_size=10, direction=DirectionEnum.ASCENDING
    )

    # Verify that only the default filterBy entries are present
    assert len(payload["query"]["filterBy"]) == 2
    assert payload["query"]["filterBy"][0]["fieldName"] == "direction"
    assert payload["query"]["filterBy"][1]["fieldName"] == "startRank"


def test_photo_album_get_url(mock_photo_library: MagicMock) -> None:
    """Tests _get_url method."""
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
    )

    url = album._get_url()

    assert url == "https://example.com/records/query?dsid=12345"


def test_photo_album_list_query_gen_with_filter(mock_photo_library: MagicMock) -> None:
    """Tests _list_query_gen method with query filter."""
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
        zone_id={"zoneName": "TestZone"},
    )

    query_filter = [
        {
            "fieldName": "custom",
            "comparator": "EQUALS",
            "fieldValue": {"type": "STRING", "value": "value"},
        }
    ]

    query = album._list_query_gen(
        offset=0,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        num_results=50,
        query_filter=query_filter,
    )

    # Verify that query filter is added to the filterBy array
    assert len(query["query"]["filterBy"]) == 3
    assert query["query"]["filterBy"][2] == query_filter[0]


def test_photo_album_list_query_gen_without_filter(
    mock_photo_library: MagicMock,
) -> None:
    """Tests _list_query_gen method without query filter."""
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        record_id="album123",
        obj_type=ObjectTypeEnum.CONTAINER,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        url="https://example.com/records/query?dsid=12345",
        zone_id={"zoneName": "TestZone"},
    )

    query = album._list_query_gen(
        offset=0,
        list_type=ListTypeEnum.CONTAINER,
        direction=DirectionEnum.ASCENDING,
        num_results=50,
        query_filter=None,
    )

    # Verify that only default filterBy entries are present
    assert len(query["query"]["filterBy"]) == 2
    assert query["query"]["filterBy"][0]["fieldName"] == "direction"
    assert query["query"]["filterBy"][1]["fieldName"] == "startRank"


def test_photo_asset_properties_and_methods() -> None:
    """Test PhotoAsset properties and methods."""

    # Prepare mock data for master and asset records
    filename = "test_photo.JPG"
    encoded_filename: str = base64.b64encode(filename.encode("utf-8")).decode("utf-8")
    now = int(datetime.now(tz=timezone.utc).timestamp() * 1000)
    master_record: dict[str, Any] = {
        "recordName": "photo_id_123",
        "fields": {
            "filenameEnc": {"value": encoded_filename},
            "resOriginalRes": {
                "value": {
                    "size": 123456,
                    "downloadURL": "http://example.com/photo.jpg",
                }
            },
            "resOriginalWidth": {"value": 1920},
            "resOriginalHeight": {"value": 1080},
            "itemType": {"value": "public.jpeg"},
            "resOriginalFileType": {"value": "public.jpeg"},
            "resJPEGThumbRes": {
                "value": {
                    "size": 1234,
                    "downloadURL": "http://example.com/thumb.jpg",
                }
            },
            "resJPEGThumbWidth": {"value": 100},
            "resJPEGThumbHeight": {"value": 50},
            "resJPEGThumbFileType": {"value": "public.jpeg"},
        },
        "recordChangeTag": "tag1",
    }
    asset_record: dict[str, Any] = {
        "fields": {
            "assetDate": {"value": now},
            "addedDate": {"value": now},
        },
        "recordName": "photo_id_123",
        "recordType": "CPLAsset",
        "zoneID": {"zoneName": "PrimarySync"},
    }

    mock_service = MagicMock()
    mock_service.service_endpoint = "https://example.com"
    mock_service.params = {"dsid": "12345"}
    mock_service.session.get.return_value = MagicMock(
        json=MagicMock(return_value={}),
        raw=MagicMock(read=MagicMock(return_value=b"response")),
    )
    mock_service.session.post.return_value = MagicMock(
        json=MagicMock(return_value={}), status_code=200
    )

    asset = PhotoAsset(mock_service, master_record, asset_record)

    # Test id
    assert asset.id == "photo_id_123"
    # Test filename
    assert asset.filename == filename
    # Test size
    assert asset.size == 123456
    # Test created and asset_date
    assert isinstance(asset.created, datetime)
    assert isinstance(asset.asset_date, datetime)
    # Test added_date
    assert isinstance(asset.added_date, datetime)
    # Test dimensions
    assert asset.dimensions == (1920, 1080)
    # Test item_type
    assert asset.item_type == "image"
    # Test is_live_photo (should be False)
    assert asset.is_live_photo is False
    # Test versions
    versions: dict[str, dict[str, Any]] = asset.versions
    assert "original" in versions
    assert "thumb" in versions
    assert versions["original"]["filename"] == filename
    assert versions["original"]["url"] == "http://example.com/photo.jpg"
    assert versions["thumb"]["url"] == "http://example.com/thumb.jpg"
    # Test download returns the mocked response
    assert asset.download(version="original") == b"response"
    # Test download with invalid version returns None
    assert asset.download(version="nonexistent") is None
    # Test delete returns a mocked response
    resp: bool = asset.delete()
    assert resp is True
    # Test __repr__
    assert repr(asset) == "<PhotoAsset: id=photo_id_123>"


def test_photo_asset_is_live_photo_true() -> None:
    """Test PhotoAsset is_live_photo property for live photo."""
    master_record: dict[str, Any] = {
        "recordName": "photo_id_456",
        "fields": {
            "filenameEnc": {
                "value": base64.b64encode(b"IMG_0001.HEIC").decode("utf-8")
            },
            "resOriginalRes": {
                "value": {
                    "size": 123456,
                    "downloadURL": "http://example.com/photo.heic",
                }
            },
            "resOriginalWidth": {"value": 4032},
            "resOriginalHeight": {"value": 3024},
            "itemType": {"value": "public.heic"},
            "resOriginalFileType": {"value": "public.heic"},
            "resOriginalVidComplFileType": {"value": "com.apple.quicktime-movie"},
            "resVidSmallRes": {
                "value": {
                    "size": 1000,
                    "downloadURL": "http://example.com/video.mov",
                }
            },
            "resVidSmallFileType": {"value": "com.apple.quicktime-movie"},
        },
        "recordChangeTag": "tag2",
    }
    asset_record: dict[str, Any] = {
        "fields": {
            "assetDate": {"value": 1700000000000},
            "addedDate": {"value": 1700000000000},
        },
        "recordName": "photo_id_456",
        "recordType": "CPLAsset",
        "zoneID": {"zoneName": "PrimarySync"},
    }
    mock_service = MagicMock()
    asset = PhotoAsset(mock_service, master_record, asset_record)
    assert asset.is_live_photo is True
    # The thumb_video version filename should end with .MOV
    thumb_video = asset.versions.get("thumb_video")
    if thumb_video:
        assert thumb_video["filename"].endswith(".MOV")


@pytest.mark.parametrize(
    "master_fields,expected_type,filename",
    [
        # itemType present and recognized
        (
            {
                "itemType": {"value": "public.jpeg"},
                "filenameEnc": {
                    "value": base64.b64encode(b"photo.JPG").decode("utf-8")
                },
            },
            "image",
            "photo.JPG",
        ),
        (
            {
                "itemType": {"value": "public.heic"},
                "filenameEnc": {
                    "value": base64.b64encode(b"photo.HEIC").decode("utf-8")
                },
            },
            "image",
            "photo.HEIC",
        ),
        (
            {
                "itemType": {"value": "com.apple.quicktime-movie"},
                "filenameEnc": {
                    "value": base64.b64encode(b"movie.MOV").decode("utf-8")
                },
            },
            "movie",
            "movie.MOV",
        ),
        # itemType missing, resOriginalFileType present and recognized
        (
            {
                "resOriginalFileType": {"value": "public.png"},
                "filenameEnc": {"value": base64.b64encode(b"img.PNG").decode("utf-8")},
            },
            "image",
            "img.PNG",
        ),
        # itemType and resOriginalFileType missing, fallback to filename extension
        (
            {
                "filenameEnc": {
                    "value": base64.b64encode(b"fallback.JPG").decode("utf-8")
                }
            },
            "image",
            "fallback.JPG",
        ),
        (
            {
                "filenameEnc": {
                    "value": base64.b64encode(b"fallback.HEIC").decode("utf-8")
                }
            },
            "image",
            "fallback.HEIC",
        ),
        # itemType and resOriginalFileType missing, filename not image, fallback to movie
        (
            {
                "filenameEnc": {
                    "value": base64.b64encode(b"fallback.avi").decode("utf-8")
                }
            },
            "movie",
            "fallback.avi",
        ),
        # itemType present but not recognized, fallback to filename extension
        (
            {
                "itemType": {"value": "unknown.type"},
                "filenameEnc": {
                    "value": base64.b64encode(b"photo.JPG").decode("utf-8")
                },
            },
            "image",
            "photo.JPG",
        ),
        (
            {
                "itemType": {"value": "unknown.type"},
                "filenameEnc": {
                    "value": base64.b64encode(b"video.avi").decode("utf-8")
                },
            },
            "movie",
            "video.avi",
        ),
    ],
)
def test_photo_asset_item_type(
    master_fields: dict[str, Any], expected_type: str, filename: str
) -> None:
    """Test PhotoAsset item_type property with various scenarios."""
    asset_record: dict[str, Any] = {
        "fields": {
            "assetDate": {"value": 1700000000000},
            "addedDate": {"value": 1700000000000},
        },
        "recordName": "photo_id_test",
        "recordType": "CPLAsset",
        "zoneID": {"zoneName": "PrimarySync"},
    }
    master_record: dict[str, Any] = {
        "recordName": "photo_id_test",
        "fields": master_fields,
        "recordChangeTag": "tag",
    }
    mock_service = MagicMock()
    asset = PhotoAsset(mock_service, master_record, asset_record)
    assert asset.filename == filename
    assert asset.item_type == expected_type


def test_shared_photo_stream_album_properties() -> None:
    """Test SharedPhotoStreamAlbum properties and methods."""

    # Setup test data
    name = "Shared Album"
    album_location = "https://shared.example.com/album/"
    album_ctag = "ctag"
    album_guid = "guid"
    owner_dsid = "owner"
    creation_date = str(
        int(datetime(2023, 1, 1, tzinfo=timezone.utc).timestamp() * 1000)
    )
    sharing_type = "owned"
    allow_contributions = True
    is_public = False
    is_web_upload_supported = True
    public_url = "https://shared.example.com/public/album"
    page_size = 50

    mock_library = MagicMock()
    album = SharedPhotoStreamAlbum(
        library=mock_library,
        name=name,
        album_location=album_location,
        album_ctag=album_ctag,
        album_guid=album_guid,
        owner_dsid=owner_dsid,
        creation_date=creation_date,
        sharing_type=sharing_type,
        allow_contributions=allow_contributions,
        is_public=is_public,
        is_web_upload_supported=is_web_upload_supported,
        public_url=public_url,
        page_size=page_size,
    )

    assert album.id == album_guid
    assert album.fullname == name
    assert album.sharing_type == sharing_type
    assert album.allow_contributions is allow_contributions
    assert album.is_public is is_public
    assert album.is_web_upload_supported is is_web_upload_supported
    assert album.public_url == public_url
    assert isinstance(album.creation_date, datetime)
    assert album._album_location == album_location
    assert album._album_ctag == album_ctag
    assert album._album_guid == album_guid
    assert album._owner_dsid == owner_dsid


def test_shared_photo_stream_album_get_payload_and_url_and_len(
    mock_photos_service: MagicMock,
) -> None:
    """Test SharedPhotoStreamAlbum _get_payload, _get_url, and _get_len."""
    mock_library = MagicMock(spec=PhotoLibrary)
    mock_library.service = mock_photos_service
    mock_photos_service.params = {"dsid": "12345"}
    mock_photos_service.session.post.return_value.json.return_value = {
        "albumassetcount": 7
    }
    mock_album = SharedPhotoStreamAlbum(
        library=mock_library,
        name="Shared Album",
        album_location="https://shared.example.com/album/",
        album_ctag="ctag",
        album_guid="guid",
        owner_dsid="owner",
        creation_date="1700000000000",
    )

    # Test _get_payload
    payload = mock_album._get_payload(
        offset=2, page_size=5, direction=DirectionEnum.ASCENDING
    )
    assert payload["albumguid"] == "guid"
    assert payload["albumctag"] == "ctag"
    assert payload["offset"] == "2"
    # limit should be offset+page_size or len(self), whichever is smaller
    # Since __len__ is not set, it will call _get_len, which returns 7
    assert payload["limit"] == str(min(2 + 5, 7))

    # Test _get_url
    url = mock_album._get_url()
    assert url.startswith("https://shared.example.com/album/webgetassets?")

    # Test _get_len
    length = mock_album._get_len()
    assert length == 7
    mock_photos_service.session.post.assert_called_with(
        "https://shared.example.com/album/webgetassetcount?dsid=12345",
        json={"albumguid": "guid"},
        headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
    )


def test_shared_photo_stream_album_delete_and_rename_are_noops() -> None:
    """Test that delete returns False and rename returns None for SharedPhotoStreamAlbum."""
    album = SharedPhotoStreamAlbum(
        library=MagicMock(),
        name="Shared Album",
        album_location="https://shared.example.com/album/",
        album_ctag="ctag",
        album_guid="guid",
        owner_dsid="owner",
        creation_date="1700000000000",
    )
    assert album.delete() is False
    assert album.rename("New Name") is None


def test_create_album_success(mock_photos_service: MagicMock) -> None:
    """Tests successful creation of an album."""
    # Mock the POST response for indexing state
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "album123",
                            "recordChangeTag": "tag123",
                            "fields": {
                                "albumNameEnc": {
                                    "value": base64.b64encode(b"My Album").decode(
                                        "utf-8"
                                    )
                                },
                                "isDeleted": {"value": False},
                            },
                        }
                    ]
                }
            )
        ),
    ]
    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    album: PhotoAlbum | None = library.create_album("My Album")
    assert album is not None
    assert album.name == "My Album"
    assert album.id == "album123"
    # Check that the correct POST was made for album creation
    expected_data = {
        "operations": [
            {
                "operationType": "create",
                "record": {
                    "recordType": "CPLAlbum",
                    "fields": {
                        "albumNameEnc": {
                            "value": base64.b64encode(
                                "My Album".encode("utf-8")
                            ).decode("utf-8"),
                        },
                        "albumType": {"value": AlbumTypeEnum.ALBUM.value},
                        "isDeleted": {"value": 0},
                        "isExpunged": {"value": 0},
                        "sortType": {"value": 1},
                        "sortAscending": {"value": 1},
                    },
                },
            }
        ],
        "zoneID": {"zoneName": "PrimarySync"},
        "atomic": True,
    }
    # The albumType value may be an enum, so just check the call was made
    assert mock_photos_service.session.post.call_count == 2
    args, kwargs = mock_photos_service.session.post.call_args
    assert "records/modify" in args[0]
    assert (
        kwargs["json"]["operations"][0]["record"]["fields"]["albumNameEnc"]["value"]
        == expected_data["operations"][0]["record"]["fields"]["albumNameEnc"]["value"]
    )


def test_create_album_returns_none_on_invalid_response(
    mock_photos_service: MagicMock,
) -> None:
    """Tests create_album returns None if _convert_record_to_album returns None."""
    # Mock the POST response for indexing state
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "album123",
                            "recordChangeTag": "tag123",
                            "fields": {
                                # Missing albumNameEnc triggers None return
                                "isDeleted": {"value": False},
                            },
                        }
                    ]
                }
            )
        ),
    ]
    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    album: PhotoAlbum | None = library.create_album("NoNameAlbum")
    assert album is None


def test_create_album_with_custom_album_type(mock_photos_service: MagicMock) -> None:
    """Tests create_album with a custom album_type."""
    # Mock the POST response for indexing state
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "album456",
                            "recordChangeTag": "tag456",
                            "fields": {
                                "albumNameEnc": {
                                    "value": base64.b64encode(b"Custom Album").decode(
                                        "utf-8"
                                    )
                                },
                                "isDeleted": {"value": False},
                            },
                        }
                    ]
                }
            )
        ),
    ]
    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )

    album: PhotoAlbum | None = library.create_album(
        "Custom Album", album_type=AlbumTypeEnum.ALBUM
    )
    assert album is not None
    assert album.name == "Custom Album"
    assert album.id == "album456"



"""Context manager to configure SSL verification for requests"""

import contextlib
import logging
import warnings
from typing import Any, Callable, Generator

import requests
from urllib3.exceptions import InsecureRequestWarning

logger: logging.Logger = logging.getLogger(__name__)


@contextlib.contextmanager
def configurable_ssl_verification(
    verify_ssl: bool = True,
    http_proxy: str | None = None,
    https_proxy: str | None = None,
) -> Generator[None, Any, None]:
    """Context manager to configure SSL verification for requests

    Warning: Setting verify_ssl=False disables certificate validation,
    making connections vulnerable to man-in-the-middle attacks.
    Only use in trusted environments for testing purposes.
    """

    # Store the original merge_environment_settings
    old_merge_environment_settings: Callable = (
        requests.Session.merge_environment_settings
    )

    def merge_environment_settings_with_config(
        self, url, proxies, stream, verify, cert
    ):
        settings = old_merge_environment_settings(
            self, url, proxies, stream, verify, cert
        )

        if not verify_ssl:
            settings["verify"] = False

        # Only set proxies if at least one is non-empty
        override_proxies: dict[str, str] = {}
        if http_proxy:
            override_proxies["http"] = http_proxy
        if https_proxy:
            override_proxies["https"] = https_proxy
        if override_proxies:
            settings["proxies"] = override_proxies
        return settings

    # Temporarily override merge_environment_settings
    requests.Session.merge_environment_settings = merge_environment_settings_with_config

    try:
        # Only catch InsecureRequestWarning if we are disabling SSL verification
        if not verify_ssl:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", InsecureRequestWarning)
                yield
        else:
            yield
    finally:
        # Restore the original merge_environment_settings
        requests.Session.merge_environment_settings = old_merge_environment_settings



"""Account family test constants."""

# Fakers
FIRST_NAME = "Quentin"
LAST_NAME = "TARANTINO"
FULL_NAME = FIRST_NAME + " " + LAST_NAME
PERSON_ID = (FIRST_NAME + LAST_NAME).lower()
PRIMARY_EMAIL = PERSON_ID + "@hotmail.fr"
APPLE_ID_EMAIL = PERSON_ID + "@me.com"
ICLOUD_ID_EMAIL = PERSON_ID + "@icloud.com"

MEMBER_1_FIRST_NAME = "John"
MEMBER_1_LAST_NAME = "TRAVOLTA"
MEMBER_1_FULL_NAME = MEMBER_1_FIRST_NAME + " " + MEMBER_1_LAST_NAME
MEMBER_1_PERSON_ID = (MEMBER_1_FIRST_NAME + MEMBER_1_LAST_NAME).lower()
MEMBER_1_APPLE_ID = MEMBER_1_PERSON_ID + "@icloud.com"

MEMBER_2_FIRST_NAME = "Uma"
MEMBER_2_LAST_NAME = "THURMAN"
MEMBER_2_FULL_NAME = MEMBER_2_FIRST_NAME + " " + MEMBER_2_LAST_NAME
MEMBER_2_PERSON_ID = (MEMBER_2_FIRST_NAME + MEMBER_2_LAST_NAME).lower()
MEMBER_2_APPLE_ID = MEMBER_2_PERSON_ID + "@outlook.fr"

FAMILY_ID = "family_" + PERSON_ID

# Data
ACCOUNT_FAMILY_WORKING = {
    "status-message": "Member of a family.",
    "familyInvitations": [],
    "outgoingTransferRequests": [],
    "isMemberOfFamily": True,
    "family": {
        "familyId": FAMILY_ID,
        "transferRequests": [],
        "invitations": [],
        "organizer": PERSON_ID,
        "members": [PERSON_ID, MEMBER_2_PERSON_ID, MEMBER_1_PERSON_ID],
        "outgoingTransferRequests": [],
        "etag": "12",
    },
    "familyMembers": [
        {
            "lastName": LAST_NAME,
            "dsid": PERSON_ID,
            "originalInvitationEmail": PRIMARY_EMAIL,
            "fullName": FULL_NAME,
            "ageClassification": "ADULT",
            "appleIdForPurchases": PRIMARY_EMAIL,
            "appleId": PRIMARY_EMAIL,
            "familyId": FAMILY_ID,
            "firstName": FIRST_NAME,
            "hasParentalPrivileges": True,
            "hasScreenTimeEnabled": False,
            "hasAskToBuyEnabled": False,
            "hasSharePurchasesEnabled": True,
            "shareMyLocationEnabledFamilyMembers": [],
            "hasShareMyLocationEnabled": True,
            "dsidForPurchases": PERSON_ID,
        },
        {
            "lastName": MEMBER_2_LAST_NAME,
            "dsid": MEMBER_2_PERSON_ID,
            "originalInvitationEmail": MEMBER_2_APPLE_ID,
            "fullName": MEMBER_2_FULL_NAME,
            "ageClassification": "ADULT",
            "appleIdForPurchases": MEMBER_2_APPLE_ID,
            "appleId": MEMBER_2_APPLE_ID,
            "familyId": FAMILY_ID,
            "firstName": MEMBER_2_FIRST_NAME,
            "hasParentalPrivileges": False,
            "hasScreenTimeEnabled": False,
            "hasAskToBuyEnabled": False,
            "hasSharePurchasesEnabled": False,
            "hasShareMyLocationEnabled": False,
            "dsidForPurchases": MEMBER_2_PERSON_ID,
        },
        {
            "lastName": MEMBER_1_LAST_NAME,
            "dsid": MEMBER_1_PERSON_ID,
            "originalInvitationEmail": MEMBER_1_APPLE_ID,
            "fullName": MEMBER_1_FULL_NAME,
            "ageClassification": "ADULT",
            "appleIdForPurchases": MEMBER_1_APPLE_ID,
            "appleId": MEMBER_1_APPLE_ID,
            "familyId": FAMILY_ID,
            "firstName": MEMBER_1_FIRST_NAME,
            "hasParentalPrivileges": False,
            "hasScreenTimeEnabled": False,
            "hasAskToBuyEnabled": False,
            "hasSharePurchasesEnabled": True,
            "hasShareMyLocationEnabled": True,
            "dsidForPurchases": MEMBER_1_PERSON_ID,
        },
    ],
    "status": 0,
    "showAddMemberButton": True,
}



"""Test constants package."""

from .const import (
    AUTHENTICATED_USER,
    REQUIRES_2FA_TOKEN,
    REQUIRES_2FA_USER,
    VALID_2FA_CODE,
    VALID_COOKIE,
    VALID_PASSWORD,
    VALID_TOKEN,
    VALID_TOKENS,
    VALID_USERS,
)
from .const_account import ACCOUNT_DEVICES_WORKING, ACCOUNT_STORAGE_WORKING
from .const_account_family import ACCOUNT_FAMILY_WORKING
from .const_drive import (
    DRIVE_FILE_DOWNLOAD_WORKING,
    DRIVE_FOLDER_WORKING,
    DRIVE_ROOT_INVALID,
    DRIVE_ROOT_WORKING,
    DRIVE_SUBFOLDER_WORKING,
    DRIVE_TRASH_DELETE_FOREVER_WORKING,
    DRIVE_TRASH_RECOVER_WORKING,
    DRIVE_TRASH_WORKING,
)
from .const_findmyiphone import FMI_FAMILY_WORKING
from .const_login import (
    AUTH_OK,
    LOGIN_2FA,
    LOGIN_WORKING,
    TRUSTED_DEVICE_1,
    TRUSTED_DEVICES,
    VERIFICATION_CODE_KO,
    VERIFICATION_CODE_OK,
)

__all__: list[str] = [
    "AUTHENTICATED_USER",
    "REQUIRES_2FA_TOKEN",
    "REQUIRES_2FA_USER",
    "VALID_2FA_CODE",
    "VALID_COOKIE",
    "VALID_TOKEN",
    "VALID_TOKENS",
    "VALID_USERS",
    "VALID_PASSWORD",
    "ACCOUNT_DEVICES_WORKING",
    "ACCOUNT_STORAGE_WORKING",
    "ACCOUNT_FAMILY_WORKING",
    "DRIVE_FILE_DOWNLOAD_WORKING",
    "DRIVE_FOLDER_WORKING",
    "DRIVE_ROOT_INVALID",
    "DRIVE_ROOT_WORKING",
    "DRIVE_SUBFOLDER_WORKING",
    "DRIVE_TRASH_DELETE_FOREVER_WORKING",
    "DRIVE_TRASH_RECOVER_WORKING",
    "DRIVE_TRASH_WORKING",
    "FMI_FAMILY_WORKING",
    "AUTH_OK",
    "LOGIN_2FA",
    "LOGIN_WORKING",
    "TRUSTED_DEVICE_1",
    "TRUSTED_DEVICES",
    "VERIFICATION_CODE_KO",
    "VERIFICATION_CODE_OK",
]



"""Tests for the SrpPassword class in pyicloud.srp_password module."""

from hashlib import sha256
from unittest.mock import patch

import pytest

from pyicloud.srp_password import SrpPassword, SrpProtocolType


def test_encode_raises_value_error_if_encrypt_info_not_set() -> None:
    """Test that encode raises ValueError if encrypt info is not set."""
    srp = SrpPassword("testpassword")
    with pytest.raises(ValueError, match="Encrypt info not set"):
        srp.encode()


def test_encode_raises_value_error_if_protocol_not_valid() -> None:
    """Test that encode raises ValueError if protocol is not valid."""

    srp = SrpPassword("testpassword")
    srp.set_encrypt_info("abc", "1", "32", None)  # type: ignore

    with pytest.raises(ValueError, match="Unsupported SrpPassword type"):
        srp.encode()


@pytest.mark.parametrize(
    "password,salt,iterations,key_length,expected_length,protocol",
    [
        ("password123", b"salty", 1000, 32, 32, SrpProtocolType.S2K),
        ("anotherpass", b"12345678", 500, 16, 16, SrpProtocolType.S2K),
        ("anotherpass", b"12345678", 500, 16, 16, SrpProtocolType.S2K_FO),
    ],
)
def test_encode_returns_correct_length(
    password: str,
    salt: bytes,
    iterations: int,
    key_length: int,
    expected_length: int,
    protocol: SrpProtocolType,
) -> None:
    """Test that encode returns bytes of the expected length."""
    srp = SrpPassword(password)
    srp.set_encrypt_info(salt, iterations, key_length, protocol)
    result: bytes = srp.encode()
    assert isinstance(result, bytes)
    assert len(result) == expected_length


def test_encode_consistency_for_same_input() -> None:
    """Test that encode returns the same result for the same input."""
    srp1 = SrpPassword("mypassword")
    srp2 = SrpPassword("mypassword")
    salt = b"abcdef"
    iterations = 1000
    key_length = 24
    srp1.set_encrypt_info(salt, iterations, key_length, SrpProtocolType.S2K)
    srp2.set_encrypt_info(salt, iterations, key_length, SrpProtocolType.S2K)
    assert srp1.encode() == srp2.encode()
    srp1.set_encrypt_info(salt, iterations, key_length, SrpProtocolType.S2K_FO)
    srp2.set_encrypt_info(salt, iterations, key_length, SrpProtocolType.S2K_FO)
    assert srp1.encode() == srp2.encode()


def test_srp_password_digest() -> None:
    """Test that the SrpPassword digest method works as expected."""
    password = "securepassword"
    salt = b"saltysalt"
    iterations = 2000
    key_length = 32

    password_digest: bytes = sha256(password.encode("utf-8")).digest()
    password_digest_hex: bytes = sha256(password.encode("utf-8")).hexdigest().encode()

    with patch("pyicloud.srp_password.pbkdf2_hmac") as mock_pbkdf2_hmac:
        srp = SrpPassword(password)
        srp.set_encrypt_info(salt, iterations, key_length, SrpProtocolType.S2K)
        _ = srp.encode()

        mock_pbkdf2_hmac.assert_called_with(
            "sha256", password_digest, salt, iterations, key_length
        )

        srp.set_encrypt_info(salt, iterations, key_length, SrpProtocolType.S2K_FO)
        _ = srp.encode()
        mock_pbkdf2_hmac.assert_called_with(
            "sha256", password_digest_hex, salt, iterations, key_length
        )



"""
Test the PyiCloudService and PyiCloudSession classes."""

# pylint: disable=protected-access

from typing import Any, List
from unittest.mock import MagicMock, mock_open, patch

import pytest
from fido2.hid import CtapHidDevice
from requests import HTTPError, Response

from pyicloud import PyiCloudService
from pyicloud.exceptions import (
    PyiCloud2SARequiredException,
    PyiCloudAcceptTermsException,
    PyiCloudAPIResponseException,
    PyiCloudFailedLoginException,
    PyiCloudServiceNotActivatedException,
    PyiCloudServiceUnavailable,
)
from pyicloud.services.calendar import CalendarService
from pyicloud.services.contacts import ContactsService
from pyicloud.services.hidemyemail import HideMyEmailService
from pyicloud.services.photos import PhotosService
from pyicloud.services.reminders import RemindersService
from pyicloud.services.ubiquity import UbiquityService
from pyicloud.session import PyiCloudSession
from pyicloud.utils import b64_encode
from tests.const import LOGIN_2FA


def test_authenticate_with_force_refresh(pyicloud_service: PyiCloudService) -> None:
    """Test the authenticate method with force_refresh=True."""
    with (
        patch("pyicloud.base.PyiCloudSession.post") as mock_post_response,
        patch("pyicloud.base.PyiCloudService._validate_token") as validate_token,
    ):
        pyicloud_service.session._data = {"session_token": "valid_token"}
        mock_post_response.json.return_value = {
            "apps": {"test_service": {"canLaunchWithOneFactor": True}},
            "status": "success",
        }
        pyicloud_service.data = {
            "apps": {"test_service": {"canLaunchWithOneFactor": True}}
        }
        validate_token = MagicMock(
            return_value={
                "status": "success",
                "dsInfo": {"hsaVersion": 1},
                "webservices": "TestWebservices",
            }
        )
        pyicloud_service._validate_token = validate_token
        pyicloud_service.authenticate(force_refresh=True, service="test_service")
        mock_post_response.assert_called_once()
        validate_token.assert_called_once()


def test_authenticate_with_missing_token(pyicloud_service: PyiCloudService) -> None:
    """Test the authenticate method with missing session_token."""
    with (
        patch("pyicloud.base.PyiCloudSession.get") as mock_get_response,
        patch("pyicloud.base.PyiCloudSession.post") as mock_post_response,
        patch.object(
            pyicloud_service,
            "_authenticate_with_token",
            side_effect=[PyiCloudFailedLoginException("a"), None],
        ) as mock_authenticate_with_token,
    ):
        mock_post_response.return_value.json.side_effect = [
            {
                "salt": "U29tZVNhbHQ=",
                "b": "U29tZUJ5dGVz",
                "c": "TestC",
                "protocol": "s2k",
                "iteration": 1000,
                "dsInfo": {"hsaVersion": 1},
                "hsaChallengeRequired": False,
                "webservices": "TestWebservices",
            },
            None,
        ]
        pyicloud_service.session.post = mock_post_response
        pyicloud_service.session._data = {}
        pyicloud_service.params = {}
        pyicloud_service.authenticate()
        assert mock_get_response.call_count == 1
        assert mock_post_response.call_count == 2
        assert mock_authenticate_with_token.call_count == 2


def test_validate_2fa_code(pyicloud_service: PyiCloudService) -> None:
    """Test the validate_2fa_code method with a valid code."""

    pyicloud_service.data = {"dsInfo": {"hsaVersion": 1}, "hsaChallengeRequired": False}

    with patch("pyicloud.base.PyiCloudSession") as mock_session:
        pyicloud_service._session = mock_session
        mock_session.data = {
            "scnt": "test_scnt",
            "session_id": "test_session_id",
            "session_token": "test_session_token",
        }

        mock_post_response = MagicMock()
        mock_post_response.status_code = 200
        mock_post_response.json.return_value = {"success": True}
        mock_session.post.return_value = mock_post_response

        assert pyicloud_service.validate_2fa_code("123456")


def test_validate_2fa_code_failure(pyicloud_service: PyiCloudService) -> None:
    """Test the validate_2fa_code method with an invalid code."""
    exception = PyiCloudAPIResponseException("Invalid code")
    exception.code = -21669
    with patch("pyicloud.base.PyiCloudSession") as mock_session:
        mock_session.post.side_effect = exception
        pyicloud_service._session = mock_session
        assert not pyicloud_service.validate_2fa_code("000000")


@patch("pyicloud.base.CtapHidDevice.list_devices", return_value=[MagicMock()])
@patch("pyicloud.base.Fido2Client")
def test_confirm_security_key_success(
    mock_fido2_client_cls, mock_list_devices, pyicloud_service: PyiCloudService
) -> None:
    """Test that the FIDO2 WebAuthn flow works"""
    rp_id = "example.com"
    challenge = "ZmFrZV9jaGFsbGVuZ2U"

    # Arrange
    pyicloud_service._submit_webauthn_assertion_response = MagicMock()
    pyicloud_service.trust_session = MagicMock()

    # Simulated WebAuthn options returned from backend
    pyicloud_service._auth_data = {
        "fsaChallenge": {
            "challenge": challenge,  # base64url(fake_challenge)
            "keyHandles": ["a2V5MQ", "a2V5Mg"],  # base64url(fake_key_ids)
            "rpId": rp_id,
        }
    }

    # Simulated FIDO2 response
    mock_response = MagicMock()
    mock_response.response = MagicMock()
    mock_response.response.client_data = b"client_data"
    mock_response.response.signature = b"signature"
    mock_response.response.authenticator_data = b"auth_data"
    mock_response.response.user_handle = b"user_handle"
    mock_response.raw_id = b"cred_id"

    mock_fido2_client = MagicMock()
    mock_fido2_client.get_assertion.return_value.get_response.return_value = (
        mock_response
    )
    mock_fido2_client_cls.return_value = mock_fido2_client

    # Act
    pyicloud_service.confirm_security_key()

    # Assert
    mock_list_devices.assert_called_once()
    mock_fido2_client.get_assertion.assert_called_once()

    # Check if data was submitted correctly
    pyicloud_service._submit_webauthn_assertion_response.assert_called_once_with(
        {
            "challenge": challenge,
            "rpId": rp_id,
            "clientData": b64_encode(mock_response.response.client_data),
            "signatureData": b64_encode(mock_response.response.signature),
            "authenticatorData": b64_encode(mock_response.response.authenticator_data),
            "userHandle": b64_encode(mock_response.response.user_handle),
            "credentialID": b64_encode(mock_response.raw_id),
        }
    )

    pyicloud_service.trust_session.assert_called_once()


def test_get_webservice_url_success(pyicloud_service: PyiCloudService) -> None:
    """Test the get_webservice_url method with a valid key."""
    pyicloud_service._webservices = {"test_key": {"url": "https://example.com"}}
    url: str = pyicloud_service.get_webservice_url("test_key")
    assert url == "https://example.com"


def test_get_webservice_url_failure(pyicloud_service: PyiCloudService) -> None:
    """Test the get_webservice_url method with an invalid key."""
    pyicloud_service._webservices = {}
    with pytest.raises(PyiCloudServiceNotActivatedException):
        pyicloud_service.get_webservice_url("invalid_key")


def test_trust_session_success(pyicloud_service: PyiCloudService) -> None:
    """Test the trust_session method with a successful response."""

    with patch("pyicloud.base.PyiCloudSession") as mock_session:
        mock_session.data = {
            "scnt": "test_scnt",
            "session_id": "test_session_id",
            "session_token": "test_session_token",
        }
        mock_session.post.return_value.json.return_value = {
            "termsUpdateNeeded": False,
            "hsaTrustedBrowser": True,
        }
        pyicloud_service._session = mock_session
        assert pyicloud_service.trust_session()


def test_trust_session_failure(pyicloud_service: PyiCloudService) -> None:
    """Test the trust_session method with a failed response."""
    with patch("pyicloud.base.PyiCloudSession") as mock_session:
        pyicloud_service._session = mock_session
        mock_session.get.side_effect = PyiCloudAPIResponseException("Trust failed")
        assert not pyicloud_service.trust_session()


def test_cookiejar_path_property(pyicloud_session: PyiCloudSession) -> None:
    """Test the cookiejar_path property."""
    path: str = pyicloud_session.cookiejar_path
    assert isinstance(path, str)


def test_session_path_property(pyicloud_session: PyiCloudSession) -> None:
    """Test the session_path property."""
    path: str = pyicloud_session.session_path
    assert isinstance(path, str)


def test_requires_2sa_property(pyicloud_service: PyiCloudService) -> None:
    """Test the requires_2sa property."""
    pyicloud_service.data = {"dsInfo": {"hsaVersion": 2}}
    assert pyicloud_service.requires_2sa


def test_requires_2fa_property(pyicloud_service: PyiCloudService) -> None:
    """Test the requires_2fa property."""
    pyicloud_service.data = LOGIN_2FA
    assert pyicloud_service.requires_2fa


def test_is_trusted_session_property(pyicloud_service: PyiCloudService) -> None:
    """Test the is_trusted_session property."""
    pyicloud_service.data = {"dsInfo": {"hsaVersion": 2}}
    assert not pyicloud_service.is_trusted_session


def test_request_success(pyicloud_service_working: PyiCloudService) -> None:
    """Test the request method with a successful response."""
    with (
        patch("requests.Session.request") as mock_request,
        patch("builtins.open", new_callable=mock_open),
        patch("os.path.exists", return_value=True),
        patch("http.cookiejar.LWPCookieJar.save") as mock_save,
        patch("http.cookiejar.LWPCookieJar.load") as mock_load,
    ):
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"success": True}
        mock_response.headers.get.return_value = "application/json"
        mock_request.return_value = mock_response
        pyicloud_session = PyiCloudSession(
            service=pyicloud_service_working,
            client_id="",
            cookie_directory="",
        )

        response: Response = pyicloud_session.request(
            "POST", "https://example.com", data={"key": "value"}
        )
        assert response.json() == {"success": True}
        assert response.headers.get("Content-Type") == "application/json"
        mock_request.assert_called_once_with(
            method="POST",
            url="https://example.com",
            data={"key": "value"},
            params=None,
            headers=None,
            cookies=None,
            files=None,
            auth=None,
            timeout=None,
            allow_redirects=True,
            proxies=None,
            hooks=None,
            stream=None,
            verify=None,
            cert=None,
            json=None,
        )
        mock_save.assert_called_once_with(
            filename="testexamplecom.cookiejar",
            ignore_discard=True,
            ignore_expires=False,
        )
        mock_load.assert_called_once_with(
            filename="testexamplecom.cookiejar",
            ignore_discard=True,
            ignore_expires=False,
        )


def test_request_failure(pyicloud_service_working: PyiCloudService) -> None:
    """Test the request method with a failure response."""

    with (
        patch("requests.Session.request") as mock_request,
        patch("builtins.open", new_callable=mock_open) as open_mock,
        patch("http.cookiejar.LWPCookieJar.save") as mock_save,
    ):
        mock_response = MagicMock()
        mock_response.status_code = 400
        mock_response.ok = False
        mock_response.json.return_value = {"error": "Bad Request"}
        mock_response.headers.get.return_value = "application/json"
        mock_request.return_value = mock_response
        pyicloud_session = PyiCloudSession(
            pyicloud_service_working, "", cookie_directory=""
        )
        with pytest.raises(PyiCloudAPIResponseException):
            pyicloud_session.request(
                "POST", "https://example.com", data={"key": "value"}
            )

        mock_request.assert_called_once_with(
            method="POST",
            url="https://example.com",
            data={"key": "value"},
            params=None,
            headers=None,
            cookies=None,
            files=None,
            auth=None,
            timeout=None,
            allow_redirects=True,
            proxies=None,
            hooks=None,
            stream=None,
            verify=None,
            cert=None,
            json=None,
        )
        mock_save.assert_called_once()
        assert open_mock.call_count == 2


def test_request_with_custom_headers(pyicloud_service_working: PyiCloudService) -> None:
    """Test the request method with custom headers."""
    with (
        patch("requests.Session.request") as mock_request,
        patch("builtins.open", new_callable=mock_open),
        patch("http.cookiejar.LWPCookieJar.save") as mock_save,
    ):
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"data": "header test"}
        mock_response.headers.get.return_value = "application/json"
        mock_request.return_value = mock_response
        pyicloud_session = PyiCloudSession(
            pyicloud_service_working, "", cookie_directory=""
        )

        response: Response = pyicloud_session.request(
            "GET",
            "https://example.com",
            headers={"Custom-Header": "Value"},
        )
        assert response.json() == {"data": "header test"}
        assert response.headers.get("Content-Type") == "application/json"
        mock_request.assert_called_once_with(
            method="GET",
            url="https://example.com",
            data=None,
            headers={"Custom-Header": "Value"},
            params=None,
            cookies=None,
            files=None,
            auth=None,
            timeout=None,
            allow_redirects=True,
            proxies=None,
            hooks=None,
            stream=None,
            verify=None,
            cert=None,
            json=None,
        )
        mock_save.assert_called_once()


def test_request_error_handling_for_response_conditions() -> None:
    """Mock the get_webservice_url to return a valid fmip_url."""
    pyicloud_service = MagicMock(spec=PyiCloudService)
    with (
        pytest.raises(PyiCloudAPIResponseException),
        patch("requests.Session.request") as mock_request,
        patch("builtins.open", new_callable=mock_open),
        patch("os.path.exists", return_value=False),
        patch("http.cookiejar.LWPCookieJar.save"),
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://fmip.example.com",
        ),
    ):
        # Mock the response with conditions that cause an error.
        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.ok = False
        mock_response.json.return_value = {"error": "Server Error"}
        mock_response.headers.get.return_value = "application/json"
        mock_request.return_value = mock_response

        pyicloud_session = PyiCloudSession(pyicloud_service, "", cookie_directory="")
        pyicloud_service.data = {"session_token": "valid_token"}

        # Use the mocked fmip_url in the request.
        pyicloud_session.request("GET", "https://fmip.example.com/path")


def test_raise_error_2sa_required(pyicloud_session: PyiCloudSession) -> None:
    """Test the _raise_error method with a 2SA required exception."""
    with (
        pytest.raises(PyiCloud2SARequiredException),
        patch("pyicloud.base.PyiCloudService.requires_2sa", return_value=True),
    ):
        pyicloud_session._raise_error(
            code=401,
            reason="Missing X-APPLE-WEBAUTH-TOKEN cookie",
            response=MagicMock(),
        )


def test_raise_error_service_not_activated(pyicloud_session: PyiCloudSession) -> None:
    """Test the _raise_error method with a service not activated exception."""
    with pytest.raises(PyiCloudServiceNotActivatedException):
        pyicloud_session._raise_error(
            code="ZONE_NOT_FOUND", reason="ServiceNotActivated", response=MagicMock()
        )


def test_raise_error_access_denied(pyicloud_session: PyiCloudSession) -> None:
    """Test the _raise_error method with an access denied exception."""
    with pytest.raises(PyiCloudAPIResponseException):
        pyicloud_session._raise_error(
            code="ACCESS_DENIED", reason="ACCESS_DENIED", response=MagicMock()
        )


def test_request_pcs_for_service_icdrs_not_disabled(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service when ICDRS is not disabled (should early return)."""
    mock_logger = MagicMock()
    pyicloud_service._session = MagicMock()
    pyicloud_service.session.post = MagicMock(
        return_value=MagicMock(json=MagicMock(return_value={"isICDRSDisabled": False}))
    )
    pyicloud_service.params = {}
    with patch("pyicloud.base.LOGGER", mock_logger):
        pyicloud_service._send_pcs_request = MagicMock()
        pyicloud_service._request_pcs_for_service("photos")
        mock_logger.warning.assert_called_once_with("ICDRS is not disabled")
        pyicloud_service._send_pcs_request.assert_not_called()


def test_request_pcs_for_service_consent_needed_and_notification_sent(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service when device consent is needed and notification is sent."""
    # First call: ICDRS disabled, device not consented
    # Second call: device consented (simulate after waiting)
    consent_states: List[dict[str, bool]] = [
        {"isICDRSDisabled": True, "isDeviceConsentedForPCS": False},
        {"isICDRSDisabled": True, "isDeviceConsentedForPCS": True},
    ]

    pyicloud_service._check_pcs_consent = MagicMock(side_effect=consent_states)
    pyicloud_service._session = MagicMock()
    pyicloud_service.params = {}
    pyicloud_service._session.post.return_value.json.side_effect = [
        {"isDeviceConsentNotificationSent": True},
        {"status": "success", "message": "ok"},
    ]
    with patch("time.sleep"):
        pyicloud_service._request_pcs_for_service("photos")
    pyicloud_service._session.post.assert_any_call(
        f"{pyicloud_service._setup_endpoint}/enableDeviceConsentForPCS",
        params=pyicloud_service.params,
    )
    # Should not raise


def test_request_pcs_for_service_consent_needed_and_notification_not_sent(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service when device consent notification is not sent (should raise)."""
    pyicloud_service._check_pcs_consent = MagicMock(
        return_value={"isICDRSDisabled": True, "isDeviceConsentedForPCS": False}
    )
    pyicloud_service._session = MagicMock()
    pyicloud_service.params = {}
    pyicloud_service._session.post.return_value.json.return_value = {
        "isDeviceConsentNotificationSent": False
    }
    with pytest.raises(
        PyiCloudAPIResponseException, match="Unable to request PCS access!"
    ):
        pyicloud_service._request_pcs_for_service("photos")


def test_request_pcs_for_service_pcs_consent_waits(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service waits for PCS consent and then proceeds."""
    # Simulate PCS consent not granted for first 2 tries, then granted
    consent_states: List[dict[str, bool]] = [
        {"isICDRSDisabled": True, "isDeviceConsentedForPCS": False},
        {"isICDRSDisabled": True, "isDeviceConsentedForPCS": False},
        {"isICDRSDisabled": True, "isDeviceConsentedForPCS": True},
    ]
    pyicloud_service._check_pcs_consent = MagicMock(side_effect=consent_states)
    pyicloud_service._session = MagicMock()
    pyicloud_service.params = {}
    pyicloud_service._session.post.return_value.json.return_value = {
        "isDeviceConsentNotificationSent": True
    }
    pyicloud_service._send_pcs_request = MagicMock(
        return_value={"status": "success", "message": "ok"}
    )
    with patch("time.sleep"):
        pyicloud_service._request_pcs_for_service("photos")
    assert pyicloud_service._send_pcs_request.called


def test_request_pcs_for_service_success_on_first_attempt(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service grants PCS access on first attempt."""
    pyicloud_service._check_pcs_consent = MagicMock(
        return_value={"isICDRSDisabled": True, "isDeviceConsentedForPCS": True}
    )
    pyicloud_service._session = MagicMock()
    pyicloud_service.params = {}
    pyicloud_service._send_pcs_request = MagicMock(
        return_value={"status": "success", "message": "ok"}
    )
    pyicloud_service._request_pcs_for_service("photos")
    pyicloud_service._send_pcs_request.assert_called_once_with(
        "photos", derived_from_user_action=True
    )


def test_request_pcs_for_service_retries_on_cookie_messages(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service retries on known cookie messages and succeeds."""
    pyicloud_service._check_pcs_consent = MagicMock(
        return_value={"isICDRSDisabled": True, "isDeviceConsentedForPCS": True}
    )
    pyicloud_service._session = MagicMock()
    pyicloud_service.params = {}
    responses: List[dict[str, str]] = [
        {"status": "error", "message": "Requested the device to upload cookies."},
        {"status": "error", "message": "Cookies not available yet on server."},
        {"status": "success", "message": "ok"},
    ]
    pyicloud_service._send_pcs_request = MagicMock(side_effect=responses)
    with patch("time.sleep"):
        pyicloud_service._request_pcs_for_service("photos")
    assert pyicloud_service._send_pcs_request.call_count == 3


def test_request_pcs_for_service_raises_on_unknown_message(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service raises on unknown PCS state message."""
    pyicloud_service._check_pcs_consent = MagicMock(
        return_value={"isICDRSDisabled": True, "isDeviceConsentedForPCS": True}
    )
    pyicloud_service._session = MagicMock()
    pyicloud_service.params = {}
    pyicloud_service._send_pcs_request = MagicMock(
        return_value={"status": "error", "message": "Some unknown error"}
    )
    mock_logger = MagicMock()

    with (
        pytest.raises(
            PyiCloudAPIResponseException, match="Unable to request PCS access!"
        ),
        patch("pyicloud.base.LOGGER", mock_logger),
    ):
        pyicloud_service._request_pcs_for_service("photos")
    mock_logger.error.assert_called()


def test_handle_accept_terms_no_terms_update_needed(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _handle_accept_terms when no terms update is needed (should do nothing)."""
    pyicloud_service.data = {"termsUpdateNeeded": False}
    login_data: dict[str, str] = {"test": "data"}
    # Should not raise or call anything
    pyicloud_service._session = MagicMock()
    pyicloud_service._accept_terms = True
    pyicloud_service._handle_accept_terms(login_data)
    pyicloud_service._session.get.assert_not_called()
    pyicloud_service._session.post.assert_not_called()


def test_handle_accept_terms_terms_update_needed_accept_terms_false(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _handle_accept_terms when terms update is needed and accept_terms is
    False (should raise)."""
    pyicloud_service.data = {"termsUpdateNeeded": True}
    pyicloud_service._accept_terms = False
    login_data: dict[str, str] = {"test": "data"}
    with pytest.raises(
        PyiCloudAcceptTermsException,
        match="You must accept the updated terms of service",
    ):
        pyicloud_service._handle_accept_terms(login_data)


def test_handle_accept_terms_terms_update_needed_accept_terms_true_success(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _handle_accept_terms when terms update is needed and accept_terms is
    True (should accept terms)."""
    pyicloud_service.data = {
        "termsUpdateNeeded": True,
        "dsInfo": {"languageCode": "en_US"},
    }
    pyicloud_service._accept_terms = True
    login_data: dict[str, str] = {"test": "data"}

    # Mock session.get and session.post
    mock_get = MagicMock()
    mock_post = MagicMock()
    pyicloud_service.session.get = mock_get
    pyicloud_service.session.post = mock_post

    # Mock getTerms response
    get_terms_response = MagicMock()
    get_terms_response.raise_for_status = MagicMock()
    get_terms_response.json.return_value = {"iCloudTerms": {"version": 42}}
    mock_get.side_effect = [get_terms_response, get_terms_response]

    # Mock accountLogin response
    post_response = MagicMock()
    post_response.raise_for_status = MagicMock()
    post_response.json.return_value = {"new": "data"}
    mock_post.return_value = post_response

    pyicloud_service._handle_accept_terms(login_data)

    # Check calls
    mock_get.assert_any_call(
        f"{pyicloud_service._setup_endpoint}/getTerms",
        params=pyicloud_service.params,
        json={"locale": "en_US"},
    )
    mock_get.assert_any_call(
        f"{pyicloud_service._setup_endpoint}/repairDone",
        params=pyicloud_service.params,
        json={"acceptedICloudTerms": 42},
    )
    mock_post.assert_called_once_with(
        f"{pyicloud_service._setup_endpoint}/accountLogin", json=login_data
    )
    assert pyicloud_service.data == {"new": "data"}


def test_handle_accept_terms_terms_update_needed_accept_terms_true_http_error(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _handle_accept_terms when terms update is needed and accept_terms is
    True but HTTP error occurs."""
    pyicloud_service.data = {
        "termsUpdateNeeded": True,
        "dsInfo": {"languageCode": "en_US"},
    }
    pyicloud_service._accept_terms = True
    login_data: dict[str, str] = {"test": "data"}

    # Mock session.get to raise HTTPError
    mock_get = MagicMock()
    pyicloud_service.session.get = mock_get
    mock_get.side_effect = HTTPError("HTTP error")

    with pytest.raises(HTTPError):
        pyicloud_service._handle_accept_terms(login_data)


def test_handle_accept_terms_terms_update_needed_accept_terms_true_post_error(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _handle_accept_terms when terms update is needed and accept_terms is
    True but POST raises HTTPError."""
    pyicloud_service.data = {
        "termsUpdateNeeded": True,
        "dsInfo": {"languageCode": "en_US"},
    }
    pyicloud_service._accept_terms = True
    login_data: dict[str, str] = {"test": "data"}

    # Mock session.get for getTerms and repairDone
    mock_get = MagicMock()
    pyicloud_service.session.get = mock_get
    get_terms_response = MagicMock()
    get_terms_response.raise_for_status = MagicMock()
    get_terms_response.json.return_value = {"iCloudTerms": {"version": 42}}
    mock_get.side_effect = [get_terms_response, get_terms_response]

    # Mock session.post to raise HTTPError
    mock_post = MagicMock()
    pyicloud_service.session.post = mock_post
    mock_post.side_effect = HTTPError("POST error")

    with pytest.raises(HTTPError):
        pyicloud_service._handle_accept_terms(login_data)


def test_validate_token_success(pyicloud_service: PyiCloudService) -> None:
    """Test _validate_token returns JSON when X-APPLE-WEBAUTH-TOKEN is present and
    request succeeds."""
    with (
        patch.object(pyicloud_service.session.cookies, "get", return_value="token"),
        patch.object(pyicloud_service.session, "post") as mock_post,
    ):
        mock_response = MagicMock()
        mock_response.json.return_value = {"status": "success"}
        mock_post.return_value = mock_response

        result = pyicloud_service._validate_token()
        assert result == {"status": "success"}
        mock_post.assert_called_once_with(
            f"{pyicloud_service._setup_endpoint}/validate", data="null"
        )


def test_validate_token_missing_cookie_raises(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _validate_token raises when X-APPLE-WEBAUTH-TOKEN cookie is missing."""
    with patch.object(pyicloud_service.session.cookies, "get", return_value=None):
        with pytest.raises(
            PyiCloudAPIResponseException, match="Missing X-APPLE-WEBAUTH-TOKEN cookie"
        ):
            pyicloud_service._validate_token()


def test_validate_token_post_raises_exception(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _validate_token raises when session.post raises PyiCloudAPIResponseException."""
    with (
        patch.object(pyicloud_service.session.cookies, "get", return_value="token"),
        patch.object(
            pyicloud_service.session,
            "post",
            side_effect=PyiCloudAPIResponseException("Invalid token"),
        ),
    ):
        with pytest.raises(PyiCloudAPIResponseException, match="Invalid token"):
            pyicloud_service._validate_token()


def test_str_and_repr(pyicloud_service: PyiCloudService) -> None:
    """Test __str__ and __repr__ methods."""
    s = str(pyicloud_service)
    r: str = repr(pyicloud_service)
    assert s.startswith("iCloud API:")
    assert r.startswith("<iCloud API:")


def test_account_name_property(pyicloud_service: PyiCloudService) -> None:
    """Test account_name property returns the correct Apple ID."""
    assert pyicloud_service.account_name == pyicloud_service._apple_id


def test_requires_2sa_true(pyicloud_service: PyiCloudService) -> None:
    """Test requires_2sa returns True when hsaVersion >= 1 and not trusted."""
    pyicloud_service.data = {
        "dsInfo": {"hsaVersion": 1},
        "hsaChallengeRequired": True,
        "hsaTrustedBrowser": False,
    }
    assert pyicloud_service.requires_2sa


def test_requires_2sa_false(pyicloud_service: PyiCloudService) -> None:
    """Test requires_2sa returns False when hsaVersion < 1."""
    pyicloud_service.data = {"dsInfo": {"hsaVersion": 0}}
    assert not pyicloud_service.requires_2sa


def test_requires_2fa_true(pyicloud_service: PyiCloudService) -> None:
    """Test requires_2fa returns True when hsaVersion == 2 and not trusted."""
    pyicloud_service.data = {
        "dsInfo": {"hsaVersion": 2},
        "hsaChallengeRequired": True,
        "hsaTrustedBrowser": False,
    }
    assert pyicloud_service.requires_2fa


def test_requires_2fa_false(pyicloud_service: PyiCloudService) -> None:
    """Test requires_2fa returns False when hsaVersion != 2."""
    pyicloud_service.data = {"dsInfo": {"hsaVersion": 1}}
    assert not pyicloud_service.requires_2fa


def test_is_trusted_session_true(pyicloud_service: PyiCloudService) -> None:
    """Test is_trusted_session returns True when hsaTrustedBrowser is True."""
    pyicloud_service.data = {"hsaTrustedBrowser": True}
    assert pyicloud_service.is_trusted_session


def test_is_trusted_session_false(pyicloud_service: PyiCloudService) -> None:
    """Test is_trusted_session returns False when hsaTrustedBrowser is False."""
    pyicloud_service.data = {"hsaTrustedBrowser": False}
    assert not pyicloud_service.is_trusted_session


def test_get_auth_headers_overrides(pyicloud_service: PyiCloudService) -> None:
    """Test _get_auth_headers applies overrides."""
    pyicloud_service.session.data["scnt"] = "test_scnt"
    pyicloud_service.session.data["session_id"] = "test_session_id"
    headers: dict[str, Any] = pyicloud_service._get_auth_headers(
        {"Extra-Header": "Value"}
    )
    assert headers["scnt"] == "test_scnt"
    assert headers["X-Apple-ID-Session-Id"] == "test_session_id"
    assert headers["Extra-Header"] == "Value"


def test_trusted_devices_calls_session_get(pyicloud_service: PyiCloudService) -> None:
    """Test trusted_devices property calls session.get and returns devices."""
    mock_response = MagicMock()
    mock_response.json.return_value = {"devices": [{"id": "device1"}]}
    pyicloud_service.session.get = MagicMock(return_value=mock_response)
    devices: list[dict[str, Any]] = pyicloud_service.trusted_devices
    assert devices == [{"id": "device1"}]
    pyicloud_service.session.get.assert_called_once()


def test_send_verification_code_success(pyicloud_service: PyiCloudService) -> None:
    """Test send_verification_code returns True on success."""
    mock_response = MagicMock()
    mock_response.json.return_value = {"success": True}
    pyicloud_service.session.post = MagicMock(return_value=mock_response)
    result = pyicloud_service.send_verification_code({"id": "device1"})
    assert result is True


def test_send_verification_code_failure(pyicloud_service: PyiCloudService) -> None:
    """Test send_verification_code returns False on failure."""
    mock_response = MagicMock()
    mock_response.json.return_value = {"success": False}
    pyicloud_service.session.post = MagicMock(return_value=mock_response)
    result: bool = pyicloud_service.send_verification_code({"id": "device1"})
    assert result is False


def test_validate_verification_code_success(pyicloud_service: PyiCloudService) -> None:
    """Test validate_verification_code returns True when code is valid."""
    pyicloud_service.session.post = MagicMock()
    pyicloud_service.trust_session = MagicMock(return_value=True)
    result: bool = pyicloud_service.validate_verification_code(
        {"id": "device1"}, "123456"
    )
    assert result is True


def test_validate_verification_code_wrong_code(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test validate_verification_code returns False on wrong code."""
    exc = PyiCloudAPIResponseException("Invalid code")
    exc.code = -21669
    pyicloud_service.session.post = MagicMock(side_effect=exc)
    result: bool = pyicloud_service.validate_verification_code(
        {"id": "device1"}, "000000"
    )
    assert result is False


def test_validate_verification_code_raises_other(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test validate_verification_code raises on unknown error."""
    exc = PyiCloudAPIResponseException("Other error")
    exc.code = 12345
    pyicloud_service.session.post = MagicMock(side_effect=exc)
    with pytest.raises(PyiCloudAPIResponseException):
        pyicloud_service.validate_verification_code({"id": "device1"}, "000000")


def test_security_key_names_returns_key_names(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test security_key_names property returns keyNames from options."""
    pyicloud_service._auth_data = {"keyNames": ["key1", "key2"]}

    assert pyicloud_service.security_key_names == ["key1", "key2"]


def test_fido2_devices_lists_devices(pyicloud_service: PyiCloudService) -> None:
    """Test fido2_devices property lists devices."""
    with patch(
        "pyicloud.base.CtapHidDevice.list_devices", return_value=[MagicMock()]
    ) as mock_list:
        devices: List[CtapHidDevice] = pyicloud_service.fido2_devices
        assert isinstance(devices, list)
        mock_list.assert_called_once()


def test_confirm_security_key_no_devices_raises(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test confirm_security_key raises if no FIDO2 devices found."""
    pyicloud_service._auth_data = {
        "fsaChallenge": {"challenge": "c", "keyHandles": [], "rpId": "rp"}
    }

    with patch("pyicloud.base.CtapHidDevice.list_devices", return_value=[]):
        with pytest.raises(RuntimeError, match="No FIDO2 devices found"):
            pyicloud_service.confirm_security_key()


def test_get_webservice_url_raises_if_missing(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test get_webservice_url raises if key missing."""
    pyicloud_service._webservices = None
    with pytest.raises(PyiCloudServiceNotActivatedException):
        pyicloud_service.get_webservice_url("missing_key")


def test_get_webservice_url_returns_url(pyicloud_service: PyiCloudService) -> None:
    """Test get_webservice_url returns correct url."""
    pyicloud_service._webservices = {"foo": {"url": "https://foo.com"}}
    assert pyicloud_service.get_webservice_url("foo") == "https://foo.com"


def test_str_returns_expected_format(pyicloud_service: PyiCloudService) -> None:
    """Test __str__ method."""
    assert str(pyicloud_service).startswith("iCloud API:")


def test_repr_returns_expected_format(pyicloud_service: PyiCloudService) -> None:
    """Test __repr__ method."""
    assert repr(pyicloud_service).startswith("<iCloud API:")


def test_account_name_property_returns_apple_id(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test account_name property returns the correct Apple ID."""
    assert pyicloud_service.account_name == pyicloud_service._apple_id


def test_hidemyemail_returns_service(pyicloud_service: PyiCloudService) -> None:
    """Test hidemyemail property returns HideMyEmailService instance."""
    mock_hme_service = MagicMock()
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://hme.example.com",
        ),
        patch(
            "pyicloud.base.HideMyEmailService", return_value=mock_hme_service
        ) as mock_hme_cls,
    ):
        pyicloud_service._hidemyemail = None
        result: HideMyEmailService = pyicloud_service.hidemyemail
        mock_hme_cls.assert_called_once_with(
            service_root="https://hme.example.com",
            session=pyicloud_service.session,
            params=pyicloud_service.params,
        )
        assert result == mock_hme_service


def test_hidemyemail_returns_cached_instance(pyicloud_service: PyiCloudService) -> None:
    """Test hidemyemail property returns cached instance if already set."""
    mock_hme_service = MagicMock()
    pyicloud_service._hidemyemail = mock_hme_service
    result: HideMyEmailService = pyicloud_service.hidemyemail
    assert result == mock_hme_service


def test_hidemyemail_raises_on_api_exception(pyicloud_service: PyiCloudService) -> None:
    """Test hidemyemail property raises PyiCloudServiceUnavailable on API exception."""
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://hme.example.com",
        ),
        patch(
            "pyicloud.base.HideMyEmailService",
            side_effect=PyiCloudAPIResponseException("error"),
        ),
    ):
        pyicloud_service._hidemyemail = None
        with pytest.raises(
            PyiCloudServiceUnavailable, match="Hide My Email service not available"
        ):
            _: HideMyEmailService = pyicloud_service.hidemyemail


def test_files_returns_service(pyicloud_service: PyiCloudService) -> None:
    """Test files property returns UbiquityService instance."""
    mock_files_service = MagicMock()
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://files.example.com",
        ),
        patch(
            "pyicloud.base.UbiquityService", return_value=mock_files_service
        ) as mock_files_cls,
    ):
        pyicloud_service._files = None
        result: UbiquityService = pyicloud_service.files
        mock_files_cls.assert_called_once_with(
            service_root="https://files.example.com",
            session=pyicloud_service.session,
            params=pyicloud_service.params,
        )
        assert result == mock_files_service


def test_files_returns_cached_instance(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test files property returns cached instance if already set."""
    mock_files_service = MagicMock()
    pyicloud_service._files = mock_files_service
    result: UbiquityService = pyicloud_service.files
    assert result == mock_files_service


def test_files_raises_on_api_exception(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test files property raises PyiCloudServiceUnavailable on API exception."""
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://files.example.com",
        ),
        patch(
            "pyicloud.base.UbiquityService",
            side_effect=PyiCloudAPIResponseException("error"),
        ),
    ):
        pyicloud_service._files = None
        with pytest.raises(
            PyiCloudServiceUnavailable, match="Files service not available"
        ):
            _: UbiquityService = pyicloud_service.files


def test_files_raises_on_account_migrated(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test files property raises specific message if Account migrated."""
    exc = PyiCloudAPIResponseException("Account migrated")
    exc.reason = "Account migrated"
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://files.example.com",
        ),
        patch(
            "pyicloud.base.UbiquityService",
            side_effect=exc,
        ),
    ):
        pyicloud_service._files = None
        with pytest.raises(
            PyiCloudServiceUnavailable,
            match="Files service not available use `api.drive` instead",
        ):
            _: UbiquityService = pyicloud_service.files


def test_photos_returns_service(pyicloud_service: PyiCloudService) -> None:
    """Test photos property returns PhotosService instance."""
    mock_photos_service = MagicMock()
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            side_effect=[
                "https://photos.example.com",
                "https://upload.example.com",
                "https://shared.example.com",
            ],
        ),
        patch(
            "pyicloud.base.PhotosService", return_value=mock_photos_service
        ) as mock_photos_cls,
        patch.object(pyicloud_service, "_request_pcs_for_service"),
    ):
        pyicloud_service._photos = None
        pyicloud_service.data = {"dsInfo": {"dsid": "12345"}}
        result: PhotosService = pyicloud_service.photos
        mock_photos_cls.assert_called_once_with(
            service_root="https://photos.example.com",
            session=pyicloud_service.session,
            params=pyicloud_service.params,
            upload_url="https://upload.example.com",
            shared_streams_url="https://shared.example.com",
        )
        assert pyicloud_service.params["dsid"] == "12345"
        assert result == mock_photos_service


def test_photos_returns_cached_instance(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test photos property returns cached instance if already set."""
    mock_photos_service = MagicMock()
    pyicloud_service._photos = mock_photos_service
    with patch.object(pyicloud_service, "_request_pcs_for_service"):
        result: PhotosService = pyicloud_service.photos
        assert result == mock_photos_service


def test_photos_raises_on_api_exception(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test photos property raises PyiCloudServiceUnavailable on API exception."""
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            side_effect=[
                "https://photos.example.com",
                "https://upload.example.com",
                "https://shared.example.com",
            ],
        ),
        patch(
            "pyicloud.base.PhotosService",
            side_effect=PyiCloudAPIResponseException("error"),
        ),
        patch.object(pyicloud_service, "_request_pcs_for_service"),
    ):
        pyicloud_service._photos = None
        pyicloud_service.data = {"dsInfo": {"dsid": "12345"}}
        with pytest.raises(
            PyiCloudServiceUnavailable, match="Photos service not available"
        ):
            _: PhotosService = pyicloud_service.photos


def test_calendar_returns_service(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test calendar property returns CalendarService instance."""
    mock_calendar_service = MagicMock()
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://calendar.example.com",
        ),
        patch(
            "pyicloud.base.CalendarService",
            return_value=mock_calendar_service,
        ) as mock_calendar_cls,
    ):
        pyicloud_service._calendar = None
        result: CalendarService = pyicloud_service.calendar
        mock_calendar_cls.assert_called_once_with(
            service_root="https://calendar.example.com",
            session=pyicloud_service.session,
            params=pyicloud_service.params,
        )
        assert result == mock_calendar_service


def test_calendar_returns_cached_instance(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test calendar property returns cached instance if already set."""
    mock_calendar_service = MagicMock()
    pyicloud_service._calendar = mock_calendar_service
    result: CalendarService = pyicloud_service.calendar
    assert result == mock_calendar_service


def test_calendar_raises_on_api_exception(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test calendar property raises PyiCloudServiceUnavailable on API exception."""
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://calendar.example.com",
        ),
        patch(
            "pyicloud.base.CalendarService",
            side_effect=PyiCloudAPIResponseException("error"),
        ),
    ):
        pyicloud_service._calendar = None
        with pytest.raises(
            PyiCloudServiceUnavailable,
            match="Calendar service not available",
        ):
            _: CalendarService = pyicloud_service.calendar


def test_contacts_returns_service(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test contacts property returns ContactsService instance."""
    mock_contacts_service = MagicMock()
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://contacts.example.com",
        ),
        patch(
            "pyicloud.base.ContactsService",
            return_value=mock_contacts_service,
        ) as mock_contacts_cls,
    ):
        pyicloud_service._contacts = None
        result: ContactsService = pyicloud_service.contacts
        mock_contacts_cls.assert_called_once_with(
            service_root="https://contacts.example.com",
            session=pyicloud_service.session,
            params=pyicloud_service.params,
        )
        assert result == mock_contacts_service


def test_contacts_returns_cached_instance(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test contacts property returns cached instance if already set."""
    mock_contacts_service = MagicMock()
    pyicloud_service._contacts = mock_contacts_service
    result: ContactsService = pyicloud_service.contacts
    assert result == mock_contacts_service


def test_contacts_raises_on_api_exception(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test contacts property raises PyiCloudServiceUnavailable on API exception."""
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://contacts.example.com",
        ),
        patch(
            "pyicloud.base.ContactsService",
            side_effect=PyiCloudAPIResponseException("error"),
        ),
    ):
        pyicloud_service._contacts = None
        with pytest.raises(
            PyiCloudServiceUnavailable,
            match="Contacts service not available",
        ):
            _: ContactsService = pyicloud_service.contacts


def test_reminders_returns_service(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test reminders property returns RemindersService instance."""
    mock_reminders_service = MagicMock()
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://reminders.example.com",
        ),
        patch(
            "pyicloud.base.RemindersService",
            return_value=mock_reminders_service,
        ) as mock_reminders_cls,
    ):
        pyicloud_service._reminders = None
        result: RemindersService = pyicloud_service.reminders
        mock_reminders_cls.assert_called_once_with(
            service_root="https://reminders.example.com",
            session=pyicloud_service.session,
            params=pyicloud_service.params,
        )
        assert result == mock_reminders_service


def test_reminders_returns_cached_instance(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test reminders property returns cached instance if already set."""
    mock_reminders_service = MagicMock()
    pyicloud_service._reminders = mock_reminders_service
    result: RemindersService = pyicloud_service.reminders
    assert result == mock_reminders_service


def test_reminders_raises_on_api_exception(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test reminders property raises PyiCloudServiceUnavailable on API exception."""
    with (
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://reminders.example.com",
        ),
        patch(
            "pyicloud.base.RemindersService",
            side_effect=PyiCloudAPIResponseException("error"),
        ),
    ):
        pyicloud_service._reminders = None
        with pytest.raises(
            PyiCloudServiceUnavailable,
            match="Reminders service not available",
        ):
            _ = pyicloud_service.reminders



"""Unit tests for UbiquityService and UbiquityNode classes."""
# pylint: disable=protected-access

from datetime import datetime
from unittest.mock import MagicMock

import pytest
from requests import Response

from pyicloud.exceptions import PyiCloudAPIResponseException, PyiCloudServiceUnavailable
from pyicloud.services.ubiquity import UbiquityNode, UbiquityService
from pyicloud.session import PyiCloudSession


def test_ubiquity_service_init() -> None:
    """Test UbiquityService initialization and exception handling."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_session.get.return_value = MagicMock(
        spec=Response, json=lambda: {"item_list": []}
    )
    params: dict[str, str] = {"dsid": "12345"}

    # Test successful initialization
    service = UbiquityService("https://example.com", mock_session, params)
    assert service.service_root == "https://example.com"
    assert service.params == params

    # Test exception handling
    mock_session.get.side_effect = PyiCloudAPIResponseException(
        code=503, reason="Service Unavailable"
    )
    with pytest.raises(PyiCloudServiceUnavailable):
        UbiquityService("https://example.com", mock_session, params)


def test_ubiquity_service_root() -> None:
    """Test the root property of UbiquityService."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_session.get.return_value = MagicMock(
        spec=Response, json=lambda: {"item_id": "0"}
    )
    service = UbiquityService("https://example.com", mock_session, {"dsid": "12345"})

    root: UbiquityNode = service.root
    assert isinstance(root, UbiquityNode)
    assert root.item_id == "0"


def test_get_node_url() -> None:
    """Test get_node_url method."""
    service = UbiquityService("https://example.com", MagicMock(), {"dsid": "12345"})
    url: str = service.get_node_url("node123")
    assert url == "https://example.com/ws/12345/item/node123"


def test_get_node() -> None:
    """Test get_node method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"item_id": "123"}
    mock_session.get.return_value = mock_response

    service = UbiquityService("https://example.com", mock_session, {"dsid": "12345"})
    node: UbiquityNode = service.get_node("123")
    assert isinstance(node, UbiquityNode)
    assert node.item_id == "123"


def test_get_children() -> None:
    """Test get_children method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {
        "item_list": [{"item_id": "1"}, {"item_id": "2"}]
    }
    mock_session.get.return_value = mock_response

    service = UbiquityService("https://example.com", mock_session, {"dsid": "12345"})
    children: list[UbiquityNode] = service.get_children("123")
    assert len(children) == 2
    assert all(isinstance(child, UbiquityNode) for child in children)


def test_ubiquity_node_properties() -> None:
    """Test UbiquityNode properties."""
    data: dict[str, str] = {
        "item_id": "123",
        "name": "Test Node",
        "type": "folder",
        "size": "1024",
        "modified": "2023-01-01T12:00:00Z",
    }
    node = UbiquityNode(MagicMock(), data)

    assert node.item_id == "123"
    assert node.name == "Test Node"
    assert node.type == "folder"
    assert node.size == 1024
    assert node.modified == datetime(2023, 1, 1, 12, 0, 0)


def test_ubiquity_node_get_children() -> None:
    """Test UbiquityNode get_children method."""
    mock_service = MagicMock(spec=UbiquityService)
    mock_service.get_children.return_value = [MagicMock(spec=UbiquityNode)]
    node = UbiquityNode(mock_service, {"item_id": "123"})

    children: list[UbiquityNode] = node.get_children()
    assert len(children) == 1
    assert isinstance(children[0], UbiquityNode)


def test_ubiquity_node_dir() -> None:
    """Test UbiquityNode dir method."""
    mock_child = MagicMock(spec=UbiquityNode)
    mock_child.name = "Child Node"
    mock_service = MagicMock(spec=UbiquityService)
    mock_service.get_children.return_value = [mock_child]
    node = UbiquityNode(mock_service, {"item_id": "123"})

    directories: list[str] = node.dir()
    assert directories == ["Child Node"]


def test_ubiquity_node_get() -> None:
    """Test UbiquityNode get method."""
    mock_child = MagicMock(spec=UbiquityNode, name="Child Node")
    mock_child.name = "Child Node"
    mock_service = MagicMock(spec=UbiquityService)
    mock_service.get_children.return_value = [mock_child]
    node = UbiquityNode(mock_service, {"item_id": "123"})

    child: UbiquityNode = node.get("Child Node")
    assert child == mock_child


def test_ubiquity_node_getitem() -> None:
    """Test UbiquityNode __getitem__ method."""
    mock_child = MagicMock(spec=UbiquityNode, name="Child Node")
    mock_child.name = "Child Node"
    mock_service = MagicMock(spec=UbiquityService)
    mock_service.get_children.return_value = [mock_child]
    node = UbiquityNode(mock_service, {"item_id": "123"})

    child: UbiquityNode = node["Child Node"]
    assert child == mock_child
